# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
import ipaddress
import ruamel.yaml
import os

from copy import deepcopy

from kubemarine import plugins
from kubemarine.core import utils



def enrich_inventory(inventory, cluster):

    # By default we use calico, but have to find it out
    # First of all we have to check is Calicon set to be installed or not
    # By default installation parameter is unset, means user did not made any decision
    if inventory["plugins"]["calico"].get("install") is None:
        # Is user defined Flannel plugin and set it to install?
        flannel_required = inventory["plugins"].get("flannel", {}).get("install", False)
        # Is user defined Canal plugin and set it to install?
        canal_required = inventory["plugins"].get("canal", {}).get("install", False)
        # If Flannel and Canal is unset or not required to install, then install Calico
        if not flannel_required and not canal_required:
            inventory["plugins"]["calico"]["install"] = True

    if inventory["plugins"]["calico"]["install"]:
        # Check if original YAML exists
        items = inventory['plugins']['calico']['installation']['procedures']
        for item in items:
            if item.get('python'):
                # create config for plugin module
                calico_original_yaml = item['python']["arguments"]["calico_original_yaml"]
                config = {
                    "source": calico_original_yaml
                }
                calico_original_yaml_path, _ = plugins.get_source_absolute_pattern(config)
                if not os.path.isfile(calico_original_yaml_path):
                    raise Exception(f"Cannot find original Calico manifest {calico_original_yaml_path}")

    return inventory


def apply_calico_yaml(cluster, calico_original_yaml, calico_yaml):
    """
    The method implements full proccessing for Calico plugin
    :param calico_original_yaml: path to original Calico manifest
    :param calico_yaml: file name of the resulting Calico manifest
    :param cluster: Cluster object
    """

    calico_yaml = os.path.basename(calico_yaml)
    destination = '/etc/kubernetes/%s' % calico_yaml

    # create config for plugin module
    config = {
        "source": calico_original_yaml,
        "destination": destination,
        "do_render": False
    }

    # get original YAML and parse it into dict of objects
    calico_original_yaml_path, _ = plugins.get_source_absolute_pattern(config)
    obj_list = load_multiple_yaml(calico_original_yaml_path, cluster)

    validate_original(cluster, obj_list)

    patched_list = []
    excluded_list = []

    # list of objects for enrichment
    objects_for_enrichment = [
            "ConfigMap_calico-config",
            "Deployment_calico-kube-controllers",
            "DaemonSet_calico-node",
            "Deployment_calico-typha",
            "Service_calico-typha", 
            "PodDisruptionBudget_calico-typha",
            "ClusterRole_calico-kube-controllers",
            "ClusterRole_calico-node"
           ]

    # enrich objects one by one
    for key in objects_for_enrichment:
        # enrich common Calico objects
        if key not in ["Service_calico-typha", "PodDisruptionBudget_calico-typha", "Deployment_calico-typha"]:
            if obj_list.get(key, ''):
                patched_list.append(key)
                obj_list = enrich_objects_fns[key](cluster, obj_list)
        else:
            # enrich 'calico-typha' objects only if it's enabled in 'cluster.yaml'
            # in other case those objects must be excluded
            str_value = utils.true_or_false(cluster.inventory['plugins']['calico']['typha']['enabled'])
            if str_value == 'false':
                obj_list.pop(key)
                excluded_list.append(key)
                cluster.log.verbose(f"The {key} has been excluded from result")
            elif str_value == 'true':
                patched_list.append(key)
                if enrich_objects_fns[key]:
                    obj_list = enrich_objects_fns[key](cluster, obj_list)
            else:
                raise Exception(f"The {key} can't be patched correctly "
                                f"plugins.calico.typha.enabled must be set in 'True' or 'False' "
                                f"as string or boolean value")

    cluster.log.verbose(f"The total number of patched objects is {len(patched_list)} "
                        f"the objects are the following: {patched_list}")
    cluster.log.verbose(f"The total number of excluded objects is {len(excluded_list)} "
                        f"the objects are the following: {excluded_list}")

    # TODO: check results 
    #validate_result()
    enriched_manifest = dump_multiple_yaml(obj_list)
    utils.dump_file(cluster, enriched_manifest, calico_yaml)
    config['source'] = io.StringIO(enriched_manifest)

    cluster.log.debug("Uploading calico manifest enriched from %s ..." % calico_original_yaml_path)
    cluster.log.debug("\tDestination: %s" % destination)

    plugins.apply_source(cluster, config)


def enrich_configmap_calico_config(cluster, obj_list):
    """
    The method implements the enrichment procedure for Calico ConfigMap
    :param cluster: Cluster object
    :param obj_list: list of objects for enrichment
    """

    key = "ConfigMap_calico-config"
    val = cluster.inventory['plugins']['calico']['mtu']
    obj_list[key]['data']['veth_mtu'] = str(val)
    cluster.log.verbose(f"The {key} has been patched in 'data.veth_mtu' with '{val}'")
    str_value = utils.true_or_false(cluster.inventory['plugins']['calico']['typha']['enabled'])
    if str_value == "true":
        val = "calico-typha"
    elif str_value == "false":
        val = "none"
    obj_list[key]['data']['typha_service_name'] = val
    cluster.log.verbose(f"The {key} has been patched in 'data.typha_service_name' with '{val}'")
    string_part = obj_list[key]['data']['cni_network_config']
    ip = cluster.inventory['services']['kubeadm']['networking']['podSubnet'].split('/')[0]
    if type(ipaddress.ip_address(ip)) is ipaddress.IPv4Address:
        val = cluster.inventory['plugins']['calico']['cni']['ipam']['ipv4']
    else:
        val = cluster.inventory['plugins']['calico']['cni']['ipam']['ipv6']
    new_string_part = string_part.replace('"type": "calico-ipam"', str(val)[:-1][1:].replace("'", "\""))
    obj_list[key]['data']['cni_network_config'] = new_string_part
    log_str = new_string_part.replace("\n", "")
    cluster.log.verbose(f"The {key} has been patched in 'data.cni_network_config' with '{log_str}'")
    
    return obj_list

def enrich_deployment_calico_kube_controllers(cluster, obj_list):
    """
    The method implements the enrichment procedure for Calico controller Deployment
    :param cluster: Cluster object
    :param obj_list: list of objects for enrichment
    """

    key = "Deployment_calico-kube-controllers"
    obj_list[key]['spec']['template']['spec']['nodeSelector'] = \
            cluster.inventory['plugins']['calico']['kube-controllers']['nodeSelector']
    for container in obj_list[key]['spec']['template']['spec']['containers']:
        if container['name'] == "calico-kube-controllers":
            num = obj_list[key]['spec']['template']['spec']['containers'].index(container)
            val = enrich_image(cluster, cluster.inventory['plugins']['calico']['kube-controllers']['image'])
            obj_list[key]['spec']['template']['spec']['containers'][num]['image'] = val
            cluster.log.verbose(f"The {key} has been patched in "
                                f"'spec.template.spec.containers.[{num}].image with '{val}'")

    return obj_list

def enrich_daemonset_calico_node(cluster, obj_list):
    """
    The method implements the enrichment procedure for Calico node DaemonSet
    :param cluster: Cluster object
    :param obj_list: list of objects for enrichment
    """

    key = "DaemonSet_calico-node"
    for container in obj_list[key]['spec']['template']['spec']['initContainers']:
        if container['name'] in ['upgrade-ipam', 'install-cni']: 
            num = obj_list[key]['spec']['template']['spec']['initContainers'].index(container)
            val = enrich_image(cluster, cluster.inventory['plugins']['calico']['cni']['image'])
            obj_list[key]['spec']['template']['spec']['initContainers'][num]['image'] = val
            cluster.log.verbose(f"The {key} has been patched in "
                                f"'spec.template.spec.initContainers.[{num}].image' with '{val}'")
        if container['name'] == "mount-bpffs":
            num = obj_list[key]['spec']['template']['spec']['initContainers'].index(container)
            val = enrich_image(cluster, cluster.inventory['plugins']['calico']['node']['image'])
            obj_list[key]['spec']['template']['spec']['initContainers'][num]['image'] = val
            cluster.log.verbose(f"The {key} has been patched in "
                                f"'spec.template.spec.initContainers.[{num}].image' with '{val}'")
        if container['name'] == "flexvol-driver":
            num = obj_list[key]['spec']['template']['spec']['initContainers'].index(container)
            val = enrich_image(cluster, cluster.inventory['plugins']['calico']['flexvol']['image'])
            obj_list[key]['spec']['template']['spec']['initContainers'][num]['image'] = val
            cluster.log.verbose(f"The {key} has been patched in "
                                f"'spec.template.spec.initContainers.[{num}].image' with '{val}'")
    for container in obj_list[key]['spec']['template']['spec']['containers']:
        if container['name'] == "calico-node":
            num = obj_list[key]['spec']['template']['spec']['containers'].index(container)
            val = enrich_image(cluster, cluster.inventory['plugins']['calico']['node']['image'])
            obj_list[key]['spec']['template']['spec']['containers'][num]['image'] = val 
            cluster.log.verbose(f"The {key} has been patched in "
                                f"'spec.template.spec.containers.[{num}].image' with '{val}'")
            ipv6_env = ['CALICO_IPV6POOL_CIDR', 'IP6', 'IP6_AUTODETECTION_METHOD', 'FELIX_IPV6SUPPORT', 
                        'CALICO_IPV6POOL_IPIP', 'CALICO_IPV6POOL_VXLAN']
            env_list = []
            for name, value in cluster.inventory['plugins']['calico']['env'].items():
                ip = cluster.inventory['services']['kubeadm']['networking']['podSubnet'].split('/')[0]
                if name not in ipv6_env and name != 'FELIX_TYPHAK8SSERVICENAME':
                    if type(value) is str:
                        env_list.append({'name': name, 'value': value})
                    elif type(value) is dict:
                        env_list.append({'name': name, 'valueFrom': value})
                    cluster.log.verbose(f"The {key} has been patched in "
                                        f"'spec.template.spec.containers.[{num}].env.{name}' with '{value}'")
                elif name in ipv6_env and type(ipaddress.ip_address(ip)) is not ipaddress.IPv4Address:
                    if type(value) is str:
                        env_list.append({'name': name, 'value': value})
                    elif type(value) is dict:
                        env_list.append({'name': name, 'valueFrom': value})
                    cluster.log.verbose(f"The {key} has been patched in "
                                        f"'spec.template.spec.containers.[{num}].env.{name}' with '{value}'")
                if utils.true_or_false(cluster.inventory['plugins']['calico']['typha']['enabled']) == "true" and \
                        name == 'FELIX_TYPHAK8SSERVICENAME':
                    env_list.append({'name': name, 'valueFrom': value})
                    cluster.log.verbose(f"The {key} has been patched in "
                                        f"'spec.template.spec.containers.[{num}].env.{name}' with '{value}'")
            i = 0
            for env in obj_list[key]['spec']['template']['spec']['containers'][num]['env']:
                for item in env_list:
                    if env['name'] == item['name']:
                        obj_list[key]['spec']['template']['spec']['containers'][num]['env'][i] = item
                i += 1

    return obj_list


def enrich_deployment_calico_typha(cluster, obj_list):
    """
    The method implements the enrichment procedure for Typha Deployment
    :param cluster: Cluster object
    :param obj_list: list of objects for enrichment
    """

    default_tolerations = [{'key': 'node.kubernetes.io/network-unavailable', 'effect': 'NoSchedule'},
                           {'key': 'node.kubernetes.io/network-unavailable', 'effect': 'NoExecute'}]

    key = "Deployment_calico-typha"
    val = cluster.inventory['plugins']['calico']['typha']['replicas']
    obj_list[key]['spec']['replicas'] = int(val)
    cluster.log.verbose(f"The {key} has been patched in 'spec.replicas' with '{val}'")
    val = cluster.inventory['plugins']['calico']['typha']['nodeSelector']
    obj_list[key]['spec']['template']['spec']['nodeSelector'] = val
    cluster.log.verbose(f"The {key} has been patched in 'spec.template.spec.nodeSelector' with '{val}'")
    for val in default_tolerations:
        obj_list[key]['spec']['template']['spec']['tolerations'].append(val)
        cluster.log.verbose(f"The {key} has been patched in 'spec.template.spec.tolerations' with '{val}'")
    for val in cluster.inventory['plugins']['calico']['typha'].get('tolerations', ''):
        obj_list[key]['spec']['template']['spec']['tolerations'].append(val)
        cluster.log.verbose(f"The {key} has been patched in 'spec.template.spec.tolerations' with '{val}'")
    for container in obj_list[key]['spec']['template']['spec']['containers']:
        if container['name'] == "calico-typha":
            num = obj_list[key]['spec']['template']['spec']['containers'].index(container)
            val = enrich_image(cluster, cluster.inventory['plugins']['calico']['typha']['image'])
            obj_list[key]['spec']['template']['spec']['containers'][num]['image'] = val
            cluster.log.verbose(f"The {key} has been patched in "
                                f"'spec.template.spec.containers.[{num}].image with '{val}'")

    return obj_list


def enrich_clusterRole_calico_kube_controllers(cluster, obj_list):
    """
    The method implements the enrichment procedure for Calico controller ClusterRole
    :param cluster: Cluster object
    :param obj_list: list of objects for enrichment
    """

    key = "ClusterRole_calico-kube-controllers"
    if cluster.inventory['rbac']['admission'] == "psp" and \
            cluster.inventory['rbac']['psp']['pod-security'] == "enabled":
        api_list = obj_list[key]['rules']
        api_list.append(psp_calico_kube_controllers)
        obj_list[key]['rules'] = api_list
        cluster.log.verbose(f"The {key} has been patched in 'rules' with '{psp_calico_kube_controllers}'")

    return obj_list


def enrich_clusterrole_calico_node(cluster, obj_list):
    """
    The method implements the enrichment procedure for Calico node ClusterRole
    :param cluster: Cluster object
    :param obj_list: list of objects for enrichment
    """

    key = "ClusterRole_calico-node"
    if cluster.inventory['rbac']['admission'] == "psp" and \
            cluster.inventory['rbac']['psp']['pod-security'] == "enabled":
        api_list = obj_list[key]['rules']
        api_list.append(psp_calico_node)
        obj_list[key]['rules'] = api_list
        cluster.log.verbose(f"The {key} has been patched in 'rules' with '{psp_calico_node}'")

    return obj_list


def validate_original(cluster, obj_list):
    """
    The method implements some validations for Calico objects
    :param cluster: Cluster object
    :param obj_list: list of objects for validation
    """

    known_objects = [
        "ConfigMap_calico-config",
        "CustomResourceDefinition_bgpconfigurations.crd.projectcalico.org",
        "CustomResourceDefinition_bgppeers.crd.projectcalico.org",
        "CustomResourceDefinition_blockaffinities.crd.projectcalico.org",
        "CustomResourceDefinition_caliconodestatuses.crd.projectcalico.org",
        "CustomResourceDefinition_clusterinformations.crd.projectcalico.org",
        "CustomResourceDefinition_felixconfigurations.crd.projectcalico.org",
        "CustomResourceDefinition_globalnetworkpolicies.crd.projectcalico.org",
        "CustomResourceDefinition_globalnetworksets.crd.projectcalico.org",
        "CustomResourceDefinition_hostendpoints.crd.projectcalico.org",
        "CustomResourceDefinition_ipamblocks.crd.projectcalico.org",
        "CustomResourceDefinition_ipamconfigs.crd.projectcalico.org",
        "CustomResourceDefinition_ipamhandles.crd.projectcalico.org",
        "CustomResourceDefinition_ippools.crd.projectcalico.org",
        "CustomResourceDefinition_ipreservations.crd.projectcalico.org",
        "CustomResourceDefinition_kubecontrollersconfigurations.crd.projectcalico.org",
        "CustomResourceDefinition_networkpolicies.crd.projectcalico.org",
        "CustomResourceDefinition_networksets.crd.projectcalico.org",
        "ClusterRole_calico-kube-controllers",
        "ClusterRoleBinding_calico-kube-controllers",
        "ClusterRole_calico-node",
        "ClusterRoleBinding_calico-node",
        "DaemonSet_calico-node",
        "ServiceAccount_calico-node",
        "Deployment_calico-kube-controllers",
        "ServiceAccount_calico-kube-controllers",
        "PodDisruptionBudget_calico-kube-controllers",
        "Deployment_calico-typha",
        "Service_calico-typha", 
        "PodDisruptionBudget_calico-typha"
    ]

    # check if there are new objects
    for key in obj_list.keys():
        if key not in known_objects:
            cluster.log.verbose(f"The current version of original yaml has a new object: {key}")

    # check if known objects were excluded
    for key in known_objects:
        if key not in list(obj_list):
            cluster.log.verbose(f"The current version of original yaml does not include"
                                f"the following object: {key}")

# TODO: implement method for validation after enrichment
# Some validation inside the objects
#def validate_result(cluster, obj_list):


def load_multiple_yaml(filepath, cluster) -> dict:
    """
    The method implements the parse YAML file that includes several YAMLs inside
    :param filepath: Path to file that should be parsed
    :return: dictionary with the 'kind' and 'name' of object as 'key' and whole YAML structure as 'value'
    """
    yaml = ruamel.yaml.YAML()
    yaml_dict = {}
    try:
        with utils.open_utf8(filepath, 'r') as stream:
            source_yamls = yaml.load_all(stream)
            for source_yaml in source_yamls:
                if source_yaml:
                    yaml_key = f"{source_yaml['kind']}_{source_yaml['metadata']['name']}"
                    # check if there is no duplication
                    if not yaml_dict.get(yaml_key, ''):
                        yaml_dict[yaml_key] = source_yaml
                    else:
                        raise Exception("ERROR: the {yaml_key} object is duplicated, please verify the original yaml")
        return yaml_dict
    except Exception as exc:
        cluster.log.error(f"Failed to load {filepath}", exc)


def dump_multiple_yaml(multi_yaml: dict) -> str:
    """
    The method implements the dumping some dictionary to the string that includes several YAMLs inside
    :param multi_yaml: dictionary with the 'kind' and 'name' of object as 'key' and whole YAML structure as 'value'
    """
    yaml = ruamel.yaml.YAML()

    with io.StringIO() as stream:
        yaml.dump_all(multi_yaml.values(), stream)
        result = stream.getvalue()

    return result


def enrich_image(cluster, image):
    """
    The method adds registry to image if it's necessary
    :param cluster: Cluster object
    :param image: particular image
    """
    if cluster.inventory['plugins']['calico']['installation'].get('registry', ''):
        if len(cluster.inventory['plugins']['calico']['installation']['registry']):
            return f"{cluster.inventory['plugins']['calico']['installation']['registry']}/{image}"

    return image


# name of objects and enrichment methods mapping
enrich_objects_fns = {
        "ConfigMap_calico-config": enrich_configmap_calico_config,
        "Deployment_calico-kube-controllers": enrich_deployment_calico_kube_controllers,
        "DaemonSet_calico-node": enrich_daemonset_calico_node,
        "Deployment_calico-typha": enrich_deployment_calico_typha,
        "ClusterRole_calico-kube-controllers": enrich_clusterRole_calico_kube_controllers,
        "ClusterRole_calico-node": enrich_clusterrole_calico_node,
        "Service_calico-typha": None, 
        "PodDisruptionBudget_calico-typha": None
}

psp_calico_kube_controllers = {
        "apiGroups": ["policy"],
        "resources": ["podsecuritypolicies"],
        "verbs":     ["use"],
        "resourceNames": ["oob-anyuid-psp"]
}

psp_calico_node = {
        "apiGroups": ["policy"],
        "resources": ["podsecuritypolicies"],
        "verbs":     ["use"],
        "resourceNames": ["oob-privileged-psp"]
}



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import io
import re
import threading
from copy import deepcopy
from typing import List, Dict, Union, Any, IO

import fabric
from invoke import UnexpectedExit

from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core import group, flow, connections
from kubemarine.core.connections import Connections
from kubemarine.core.group import NodeGroup, NodeGroupResult, _GenericResult, _HostToResult
from kubemarine.core.resources import DynamicResources

ShellResult = Dict[str, Union[NodeGroupResult, Any]]


class FakeShell:
    def __init__(self):
        self.results: Dict[str, List[ShellResult]] = {}
        self.history: Dict[str, List[ShellResult]] = {}
        self._lock = threading.Lock()

    def __deepcopy__(self, memodict={}):
        cls = self.__class__
        result = cls.__new__(cls)
        memodict[id(self)] = result
        result.results = deepcopy(self.results, memodict)
        result.history = deepcopy(self.history, memodict)
        result._lock = threading.Lock()
        return result

    def reset(self):
        self.results = {}
        self.history = {}

    def add(self, results: Union[NodeGroupResult, _HostToResult], do_type, args, usage_limit=0):
        args.sort()

        for host, result in results.items():
            host = host.host if isinstance(host, fabric.connection.Connection) else host
            result = {
                'result': result,
                'do_type': do_type,
                'args': args,
                'used_times': 0
            }

            if usage_limit > 0:
                result['usage_limit'] = usage_limit

            self.results.setdefault(host, []).append(result)

    def find(self, host: str, do_type, args, kwargs) -> _GenericResult:
        # TODO: Support kwargs
        with self._lock:
            if isinstance(args, tuple):
                args = list(args)
            results = self.results.get(host, [])
            for i, item in enumerate(results):
                if item['do_type'] == do_type and item['args'] == args:
                    history_found = any(history_item is item for history_item in self.history.get(host, []))
                    if not history_found:
                        self.history.setdefault(host, []).append(item)

                    item['used_times'] += 1
                    if item.get('usage_limit') is not None:
                        item['usage_limit'] -= 1
                        if item['usage_limit'] < 1:
                            del results[i]
                    return item['result']
            return None

    # covered by test.test_demo.TestFakeShell.test_calculate_calls
    def history_find(self, host: str, do_type, args):
        # TODO: Support kwargs
        result = []
        if isinstance(args, tuple):
            args = list(args)
        for item in self.history.get(host, []):
            if item['do_type'] == do_type and item['args'] == args:
                result.append(item)
        return result

    def is_called_each(self, hosts: List[str], do_type: str, args: list) -> bool:
        return all((self.is_called(host, do_type, args) for host in hosts))

    def is_called(self, host: str, do_type: str, args: list) -> bool:
        """
        Returns true if the specified command has already been executed in FakeShell for the specified connection.
        :param host: host to check, for which the desirable command should have been executed.
        :param do_type: The type of required command
        :param args: Required command arguments
        :return: Boolean
        """
        found_entries = self.history_find(host, do_type, args)
        return sum(found_entry['used_times'] for found_entry in found_entries) > 0


class FakeFS:
    def __init__(self):
        self.storage: Dict[str, Dict[str, str]] = {}
        self._lock = threading.Lock()

    def __deepcopy__(self, memodict={}):
        cls = self.__class__
        result = cls.__new__(cls)
        memodict[id(self)] = result
        result.storage = deepcopy(self.storage, memodict)
        result._lock = threading.Lock()
        return result

    def reset(self):
        self.storage = {}

    def reset_host(self, host):
        self.storage[host] = {}

    # covered by test.test_demo.TestFakeFS.test_put_string
    # covered by test.test_demo.TestFakeFS.test_put_stringio
    # covered by test.test_demo.TestFakeFS.test_write_file_to_cluster
    def write(self, host, filename, data):
        with self._lock:
            if isinstance(data, io.BytesIO):
                data = data.getvalue().decode('utf-8')
            elif isinstance(data, str):
                # this is for self-testing purpose
                pass
            elif isinstance(data, io.IOBase):
                data = data.read().decode('utf-8')
            else:
                raise ValueError("Unsupported data type " + str(type(data)))
            if self.storage.get(host) is None:
                self.storage[host] = {}
            self.storage[host][filename] = data

    # covered by test.test_demo.TestFakeFS.test_put_string
    # covered by test.test_demo.TestFakeFS.test_get_nonexistent
    def read(self, host, filename):
        return self.storage.get(host, {}).get(filename)

    # covered by test.test_demo.TestFakeFS.test_write_file_to_cluster
    def read_all(self, hosts: List[str], filename):
        result = {}
        for host in hosts:
            result[host] = self.read(host, filename)
        return result

    def ls(self, host, path):
        for _path in list(self.storage.get(host, {}).keys()):
            # TODO
            pass

    def rm(self, host, path):
        for _path in list(self.storage.get(host, {}).keys()):
            if path in _path:
                del self.storage[host][_path]


class FakeKubernetesCluster(KubernetesCluster):

    def __init__(self, *args, **kwargs):
        self.fake_shell = kwargs.pop("fake_shell", FakeShell())
        self.fake_fs = kwargs.pop("fake_fs", FakeFS())
        super().__init__(*args, **kwargs)
        self._connection_pool = FakeConnectionPool(self)

    def make_group(self, ips) -> NodeGroup:
        nodegroup = super().make_group(ips)
        return FakeNodeGroup(nodegroup.nodes, self)

    def dump_finalized_inventory(self):
        return

    def preserve_inventory(self):
        return


class FakeResources(DynamicResources):
    def __init__(self, context, raw_inventory: dict, procedure_inventory: dict = None,
                 cluster: KubernetesCluster = None,
                 fake_shell: FakeShell = None, fake_fs: FakeFS = None):
        super().__init__(context, True)
        self.inventory_filepath = None
        self.procedure_inventory_filepath = None
        self._raw_inventory = raw_inventory
        self._formatted_inventory = raw_inventory
        self._procedure_inventory = procedure_inventory
        self._cluster = cluster
        if cluster:
            self._logger = cluster.log
        self._fake_shell = fake_shell if fake_shell else FakeShell()
        self._fake_fs = fake_fs if fake_fs else FakeFS()

    def _new_cluster_instance(self, context: dict):
        return FakeKubernetesCluster(self.raw_inventory(), context,
                                     procedure_inventory=self.procedure_inventory(),
                                     logger=self.logger(),
                                     fake_shell=self._fake_shell, fake_fs=self._fake_fs)


class FakeConnection(fabric.connection.Connection):

    def __init__(self, ip, cluster: FakeKubernetesCluster, **kw):
        super().__init__(ip, **kw)
        self.fake_shell = cluster.fake_shell
        self.fake_fs = cluster.fake_fs

        command_sep = r'[=\-_]{32}'
        sep_symbol = r'\&\&|;'
        final_sep = rf" ({sep_symbol}) " \
                    rf"echo \"({command_sep})\" \1 " \
                    rf"echo \$\? \1 " \
                    rf"echo \"\2\" \1 " \
                    rf"echo \"\2\" 1>\&2 \1 (sudo )?"

        self.separator_ptrn = re.compile(final_sep)

    def __setattr__(self, key, value):
        # fabric Connection has special handling of this method. Call default behaviour for custom attributes.
        if key in ('fake_shell', 'fake_fs', 'separator_ptrn'):
            return object.__setattr__(self, key, value)
        super().__setattr__(key, value)

    def run(self, command, **kwargs) -> fabric.runners.Result:
        return self._do("run", command, **kwargs)

    def sudo(self, command, **kwargs) -> fabric.runners.Result:
        return self._do("sudo", command, **kwargs)

    # not implemented
    def get(self, *args, **kwargs):
        return self._do("get", *args, **kwargs)

    def put(self, *args, **kwargs):
        return self._do("put", *args, **kwargs)

    def _do(self, do_type, *args, **kwargs) -> fabric.runners.Result:
        if do_type in ['sudo', 'run']:
            # start fake execution of commands
            command = list(args)[0]
            commands, sep_symbol, command_sep = self._split_command(do_type, command)

            stdout = ""
            stderr = ""
            prev_exited = None
            i = 0
            for command in commands:
                found_result = self.fake_shell.find(self.host, do_type, [command], kwargs)

                if found_result is None:
                    raise Exception('Fake result not found for requested action type \'%s\' and command %s' % (do_type, [command]))

                if isinstance(found_result, Exception):
                    if i > 0:
                        raise ValueError("Exception can be thrown only for the whole command")
                    else:
                        raise found_result

                if i > 0:
                    stdout += command_sep + '\n' + str(prev_exited) + '\n' + command_sep + '\n'
                    stderr += command_sep + '\n'
                i += 1

                stdout += found_result.stdout
                stderr += found_result.stderr
                prev_exited = found_result.exited
                if prev_exited != 0:
                    if sep_symbol == ';':
                        prev_exited = 0  # todo bug of RemoteExecutor
                    else:
                        # stop fake execution
                        break

            final_res = fabric.runners.Result(stdout=stdout, stderr=stderr, exited=prev_exited, connection=self)
            if prev_exited == 0 or kwargs.get('warn', False):
                return final_res

            raise UnexpectedExit(final_res)

        elif do_type == 'put':
            # It should return fabric.transfer.Result, but currently returns None.
            # Transfer Result is currently never handled.
            return self.fake_fs.write(self.host, args[1], args[0])

        raise Exception('Unsupported do type')

    def _split_command(self, do_type, command: str):
        """
        This is a reverse operation to the RemoteExecutor#_merge_actions
        """
        tokens = self.separator_ptrn.split(command)
        commands = []
        i = 0
        sep_symbol = None
        command_sep = None
        while i < len(tokens):
            commands.append(tokens[i])
            i += 1
            if i < len(tokens):
                sep_symbol = self._compare_and_return(sep_symbol, tokens[i], "Separator symbols are not equal")
                command_sep = self._compare_and_return(command_sep, tokens[i + 1], "Command separators are not equal")
                resolved_do_type = "sudo" if tokens[i + 2] == 'sudo ' else "run"
                self._compare_and_return(do_type, resolved_do_type, "Do types are not equal")
                i += 3

        return commands, sep_symbol, command_sep

    def _compare_and_return(self, one, another, msg):
        if one is None or one == another:
            return another
        else:
            raise ValueError(msg)

    def close(self):
        pass


class FakeNodeGroup(group.NodeGroup):

    def __init__(self, connections: Connections, cluster_: FakeKubernetesCluster):
        super().__init__(connections, cluster_)

    def get_local_file_sha1(self, filename):
        return '0'

    def get_remote_file_sha1(self, filename):
        return {host: '1' for host in self.nodes.keys()}

    def _put(self, local_stream: IO, remote_file: str, **kwargs):
        kwargs.pop("sudo", None)
        kwargs.pop("backup", None)
        kwargs.pop("immutable", None)
        super()._put(local_stream, remote_file, **kwargs)


class FakeConnectionPool(connections.ConnectionPool):
    def __init__(self, cluster: FakeKubernetesCluster):
        super().__init__(cluster)
        self.cluster = cluster

    def _create_connection_from_details(self, ip: str, conn_details: dict, gateway=None, inline_ssh_env=True):
        return FakeConnection(
            ip, self.cluster,
            user=conn_details.get('username', self._env.globals['connection']['defaults']['username']),
            port=conn_details.get('connection_port', self._env.globals['connection']['defaults']['port'])
        )


def create_silent_context(args: list = None, parser: argparse.ArgumentParser = None, procedure: str = None):
    args = list(args) if args else []
    # todo probably increase logging level to get rid of spam in logs.
    if '--disable-dump' not in args:
        args.append('--disable-dump')

    if parser is None:
        parser = flow.new_common_parser("Help text")
    context = flow.create_context(parser, args, procedure=procedure)
    del context['execution_arguments']['ansible_inventory_location']
    context['preserve_inventory'] = False

    return context


def new_cluster(inventory, procedure_inventory=None, context: dict = None,
                fake=True) -> Union[KubernetesCluster, FakeKubernetesCluster]:
    if context is None:
        context = create_silent_context()

    nodes_context = generate_nodes_context(inventory)
    nodes_context.update(context['nodes'])
    context['nodes'] = nodes_context

    # It is possible to disable FakeCluster and create real cluster Object for some business case
    if fake:
        cluster = FakeKubernetesCluster(inventory, context, procedure_inventory=procedure_inventory)
    else:
        cluster = KubernetesCluster(inventory, context, procedure_inventory=procedure_inventory)

    cluster.enrich()
    return cluster


def generate_nodes_context(inventory: dict, os_name='centos', os_version='7.9', net_interface='eth0') -> dict:
    os_family = None

    if os_name in ['centos', 'rhel']:
        os_family = 'rhel'
    elif os_name in ['ubuntu', 'debian']:
        os_family = 'debian'

    context = {}
    for node in inventory['nodes']:
        node_context = {
            'access': {
                'online': True,
                'accessible': True,
                'sudo': 'Root'
            },
            'active_interface': net_interface,
            'os': {
                'name': os_name,
                'family': os_family,
                'version': os_version
            }
        }
        connect_to = node['internal_address']
        if node.get('address'):
            connect_to = node['address']
        context[connect_to] = node_context

    return context


def generate_inventory(balancer=1, master=1, worker=1, keepalived=0, haproxy_mntc=0):
    inventory: dict = {
        'node_defaults': {
            'keyfile': '/dev/null',
            'username': 'anonymous'
        },
        'nodes': [],
        'services': {
            'cri': {}
        },
        'cluster_name': 'k8s.fake.local'
    }

    id_roles_map = {}

    for role_name in ['balancer', 'master', 'worker']:

        item = locals()[role_name]

        if isinstance(item, int):
            ids = []
            if item > 0:
                for i in range(0, item):
                    ids.append('%s-%s' % (role_name, i + 1))
            item = ids

        if item:
            for id_ in item:
                roles = id_roles_map.get(id_)
                if roles is None:
                    roles = []
                roles.append(role_name)
                id_roles_map[id_] = roles

    ip_i = 0

    for id_, roles in id_roles_map.items():
        ip_i = ip_i + 1
        if "master" in roles and worker == 0:
            roles.append('worker')
        inventory['nodes'].append({
            'name': id_,
            'address': '10.101.1.%s' % ip_i,
            'internal_address': '192.168.0.%s' % ip_i,
            'roles': roles
        })

    ip_i = 0
    vrrp_ips = []

    if isinstance(keepalived, list):
        vrrp_ips.append(deepcopy(keepalived))
    elif isinstance(keepalived, int) and keepalived > 0:
        for _ in range(keepalived):
            ip_i = ip_i + 1
            vrrp_ips.append('10.101.2.%s' % ip_i)

    if isinstance(haproxy_mntc, list):
        vrrp_ips.append(deepcopy(haproxy_mntc))
    elif isinstance(haproxy_mntc, int) and haproxy_mntc > 0:
        for _ in range(haproxy_mntc):
            ip_i = ip_i + 1
            vrrp_ips.append({
                'ip': '10.101.2.%s' % ip_i,
                'params': {
                    'maintenance-type': 'not bind'
                }
            })

    if haproxy_mntc:
        inventory['services'] = {
            'loadbalancer' : {
                'haproxy': {
                    'maintenance_mode': True
                }
            }
        }

    inventory['vrrp_ips'] = vrrp_ips

    return inventory


def create_exception_result(group_: NodeGroup, exception: Exception) -> NodeGroupResult:
    return NodeGroupResult(group_.cluster, create_hosts_exception_result(group_.get_hosts(), exception))


def create_hosts_exception_result(hosts: List[str], exception: Exception) -> _HostToResult:
    return {host: exception for host in hosts}


def create_nodegroup_result(group_: NodeGroup, stdout='', stderr='', code=0) -> NodeGroupResult:
    return NodeGroupResult(group_.cluster, create_hosts_result(group_.get_hosts(), stdout, stderr, code))


def create_hosts_result(hosts: List[str], stdout='', stderr='', code=0) -> _HostToResult:
    # each host should have its own result instance.
    return {host: create_result(stdout, stderr, code) for host in hosts}


def create_result(stdout='', stderr='', code=0) -> _GenericResult:
    # connection will be later replaced to fake
    return fabric.runners.Result(stdout=stdout, stderr=stderr, exited=code, connection=None)


def empty_action(*args, **kwargs) -> None:
    """
    A dummy method that does nothing
    :return: None
    """
    pass


def new_scheme(scheme: dict, role: str, number: int):
    scheme = deepcopy(scheme)
    scheme[role] = number
    return scheme


FULLHA = {'balancer': 1, 'master': 3, 'worker': 3}
FULLHA_KEEPALIVED = {'balancer': 2, 'master': 3, 'worker': 3, 'keepalived': 1}
FULLHA_NOBALANCERS = {'balancer': 0, 'master': 3, 'worker': 3}
ALLINONE = {'master': 1, 'balancer': ['master-1'], 'worker': ['master-1'], 'keepalived': 1}
MINIHA = {'master': 3}
MINIHA_KEEPALIVED = {'master': 3, 'balancer': ['master-1', 'master-2', 'master-3'],
                     'worker': ['master-1', 'master-2', 'master-3'], 'keepalived': 1}
NON_HA_BALANCER = {'balancer': 1, 'master': 3, 'worker': ['master-1', 'master-2', 'master-3']}



#!/usr/bin/env python3
# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import OrderedDict
from copy import deepcopy
from io import StringIO

import toml

from distutils.util import strtobool

from kubemarine.core.action import Action
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.resources import DynamicResources
from kubemarine.core.yaml_merger import default_merger
from kubemarine.core import flow
from kubemarine.procedures import install
from kubemarine import kubernetes, plugins
from itertools import chain
from kubemarine.core import utils
from kubemarine.core.executor import RemoteExecutor






def system_prepare_thirdparties(cluster):
    if not cluster.inventory['services'].get('thirdparties', {}):
        cluster.log.debug("Skipped - no thirdparties defined in config file")
        return

    install.system_prepare_thirdparties(cluster)


def prepull_images(cluster):
    cluster.log.debug("Prepulling Kubernetes images...")
    fix_cri_socket(cluster)
    upgrade_group = kubernetes.get_group_for_upgrade(cluster)
    upgrade_group.call(kubernetes.images_grouped_prepull)


def kubernetes_upgrade(cluster):
    version = cluster.inventory["services"]["kubeadm"]["kubernetesVersion"]
    upgrade_group = kubernetes.get_group_for_upgrade(cluster)

    drain_timeout = cluster.procedure_inventory.get('drain_timeout')
    grace_period = cluster.procedure_inventory.get('grace_period')

    kubernetes.upgrade_first_control_plane(version, upgrade_group, cluster,
                                    drain_timeout=drain_timeout, grace_period=grace_period)

    # After first control-plane upgrade is finished we may loose our CoreDNS changes.
    # Thus, we need to re-apply our CoreDNS changes immediately after first control-plane upgrade.
    install.deploy_coredns(cluster)

    kubernetes.upgrade_other_control_planes(version, upgrade_group, cluster,
                                     drain_timeout=drain_timeout, grace_period=grace_period)
    if cluster.nodes.get('worker', []):
        kubernetes.upgrade_workers(version, upgrade_group, cluster,
                                   drain_timeout=drain_timeout, grace_period=grace_period)

    cluster.nodes['control-plane'].get_first_member().sudo('rm -f /etc/kubernetes/nodes-k8s-versions.txt')
    cluster.context['cached_nodes_versions_cleaned'] = True


def kubernetes_cleanup_nodes_versions(cluster):
    if not cluster.context.get('cached_nodes_versions_cleaned', False):
        cluster.log.verbose('Cached nodes versions required')
        cluster.nodes['control-plane'].get_first_member().sudo('rm -f /etc/kubernetes/nodes-k8s-versions.txt')
    else:
        cluster.log.verbose('Cached nodes versions already cleaned')
    kubernetes_apply_taints(cluster)


def upgrade_packages(cluster: KubernetesCluster):
    upgrade_version = cluster.context["upgrade_version"]

    packages = cluster.procedure_inventory.get(upgrade_version, {}).get("packages", {})
    if packages.get("install") is not None or packages.get("upgrade") is not None or packages.get("remove") is not None:
        install.manage_custom_packages(cluster.nodes['all'])


def upgrade_plugins(cluster):
    upgrade_version = cluster.context["upgrade_version"]

    # upgrade_candidates is a source of upgradeable plugins, not list of plugins to upgrade.
    # Some plugins from upgrade_candidates will not be upgraded, because they have "install: false"
    upgrade_candidates = {}
    defined_plugins = cluster.procedure_inventory.get(upgrade_version, {}).get("plugins", {}).keys()
    for plugin in chain(defined_plugins, plugins.oob_plugins):
        # TODO: use only OOB plugins that have changed version so that we do not perform redundant installations
        upgrade_candidates[plugin] = cluster.inventory["plugins"][plugin]

    plugins.install(cluster, upgrade_candidates)


def upgrade_containerd(cluster: KubernetesCluster):
    """
        This function fixes the incorrect version of pause during the cluster update procedure
    """


    cri = cluster.inventory["services"]["cri"]['containerRuntime']
    if cri == 'containerd':
        path = 'plugins."io.containerd.grpc.v1.cri"'
        target_kubernetes_version = cluster.context["upgrade_version"]
        index_pos = target_kubernetes_version.rfind(".")
        target_kubernetes_version = target_kubernetes_version[:index_pos]
        pause_version = cluster.globals['compatibility_map']['software']['pause'][target_kubernetes_version]['version']
        if not cluster.inventory["services"]["cri"]['containerdConfig'].get(path, False):
            return
        last_pause_version = cluster.inventory["services"]["cri"]['containerdConfig'][path]["sandbox_image"].split(":")[2]
        if last_pause_version != pause_version:
            sandbox = cluster.inventory["services"]["cri"]['containerdConfig'][path]["sandbox_image"]
            param_begin_pos = sandbox.rfind(":")
            sandbox = sandbox[:param_begin_pos] + ":" + str(pause_version)
            cluster.inventory["services"]["cri"]['containerdConfig'][path]["sandbox_image"] = sandbox
            config_string = ""
            containerd_config = cluster.inventory["services"]["cri"]['containerdConfig']
            runc_options_path = 'plugins."io.containerd.grpc.v1.cri".containerd.runtimes.runc.options'
            if not isinstance(containerd_config[runc_options_path]['SystemdCgroup'], bool):
                containerd_config[runc_options_path]['SystemdCgroup'] = \
                    bool(strtobool(containerd_config[runc_options_path]['SystemdCgroup']))
            for key, value in containerd_config.items():
                # first we process all "simple" `key: value` pairs
                if not isinstance(value, dict):
                    config_string += f"{toml.dumps({key: value})}"
            for key, value in containerd_config.items():
                # next we process all "complex" `key: dict_value` pairs, representing named sections
                if isinstance(value, dict):
                    config_string += f"\n[{key}]\n{toml.dumps(value)}"
            utils.dump_file(cluster, config_string, 'containerd-config.toml')
            with RemoteExecutor(cluster) as exe:
                for node in cluster.nodes['control-plane'].include_group(cluster.nodes.get('worker')).get_ordered_members_list(
                        provide_node_configs=True):
                    os_specific_associations = cluster.get_associations_for_node(node['connect_to'], 'containerd')
                    node['connection'].put(StringIO(config_string), os_specific_associations['config_location'],
                                           backup=True,
                                           sudo=True, mkdir=True)
                    node['connection'].sudo(f"sudo systemctl restart {os_specific_associations['service_name']} && "
                                            f"systemctl status {os_specific_associations['service_name']}")
            return exe.get_last_results_str()


tasks = OrderedDict({
    "verify_upgrade_versions": kubernetes.verify_upgrade_versions,
    "thirdparties": system_prepare_thirdparties,
    "prepull_images": prepull_images,
    "configure_policy": install.system_prepare_policy,
    "kubernetes": kubernetes_upgrade,
    "kubernetes_cleanup": kubernetes_cleanup_nodes_versions,
    "packages": upgrade_packages,
    "upgrade_containerd": upgrade_containerd,
    "plugins": upgrade_plugins,
    "overview": install.overview

})


def upgrade_finalize_inventory(cluster, inventory):
    if cluster.context.get("initial_procedure") != "upgrade":
        return inventory
    upgrade_version = cluster.context.get("upgrade_version")

    inventory.setdefault("services", {}).setdefault("kubeadm", {})['kubernetesVersion'] = upgrade_version

    # if thirdparties was not defined in procedure.yaml,
    # then no need to forcibly place them: user may want to use default
    if cluster.procedure_inventory.get(upgrade_version, {}).get('thirdparties'):
        inventory['services']['thirdparties'] = cluster.procedure_inventory[upgrade_version]['thirdparties']

    if cluster.procedure_inventory.get(upgrade_version, {}).get("plugins"):
        inventory.setdefault("plugins", {})
        default_merger.merge(inventory["plugins"], cluster.procedure_inventory[upgrade_version]["plugins"])

    if cluster.procedure_inventory.get(upgrade_version, {}).get("packages"):
        inventory['services'].setdefault("packages", {})
        packages_section = deepcopy(cluster.procedure_inventory[upgrade_version]["packages"])
        for _type in ['install', 'upgrade', 'remove']:
            packages = packages_section.pop(_type, None)
            if packages is None:
                continue
            if isinstance(packages, list):
                packages = {'include': packages}

            packages_list = inventory['services']['packages'].get(_type)
            if isinstance(packages_list, list):
                inventory['services']['packages'][_type] = {
                    'include': packages_list
                }
            default_merger.merge(inventory["services"]["packages"].setdefault(_type, {}), packages)
        # Despite we enrich OS specific section inside system.enrich_upgrade_inventory,
        # we still merge global associations section because it has priority during enrichment.
        default_merger.merge(inventory["services"]["packages"], packages_section)

    return inventory


class UpgradeAction(Action):
    def __init__(self, upgrade_version: str):
        super().__init__('upgrade to ' + upgrade_version, recreate_inventory=True)
        self.upgrade_version = upgrade_version

    def run(self, res: DynamicResources):
        flow.run_tasks(res, tasks)
        res.make_final_inventory()

    def prepare_context(self, context: dict):
        context['upgrade_version'] = self.upgrade_version
        context['dump_filename_prefix'] = self.upgrade_version


def main(cli_arguments=None):
    cli_help = '''
    Script for automated upgrade of the entire Kubernetes cluster to a new version.

    How to use:

    '''

    parser = flow.new_procedure_parser(cli_help, tasks=tasks)

    context = flow.create_context(parser, cli_arguments, procedure='upgrade')
    args = context['execution_arguments']
    resources = DynamicResources(context)

    upgrade_plan = verify_upgrade_plan(resources.procedure_inventory().get('upgrade_plan'))
    verification_version_result = kubernetes.verify_target_version(upgrade_plan[-1])

    if (args['tasks'] or args['exclude']) and len(upgrade_plan) > 1:
        raise Exception("Usage of '--tasks' and '--exclude' is not allowed when upgrading to more than one version")

    # todo inventory is preserved few times, probably need to preserve it once instead.
    actions = [UpgradeAction(version) for version in upgrade_plan]
    flow.run_actions(resources, actions)

    if verification_version_result:
        print(verification_version_result)


def verify_upgrade_plan(upgrade_plan):
    if not upgrade_plan:
        raise Exception('Upgrade plan is not specified or empty')

    upgrade_plan.sort()

    previous_version = None
    for i in range(0, len(upgrade_plan)):
        version = upgrade_plan[i]
        if version == 'v1.24.0':
            raise Exception('Attempt to upgrade to an unstable version of kubernetes')
        if previous_version is not None:
            kubernetes.test_version_upgrade_possible(previous_version, version)
        previous_version = version

    print('Loaded upgrade plan: current ⭢', ' ⭢ '.join(upgrade_plan))

    return upgrade_plan


def fix_cri_socket(cluster):
    """
    This method fixs the issue with 'kubeadm.alpha.kubernetes.io/cri-socket' node annotation
    and delete the docker socket if it exists
    """

    if cluster.inventory["services"]["cri"]["containerRuntime"] == "containerd":
        control_plane = cluster.nodes["control-plane"].get_first_member(provide_node_configs=True)
        control_plane["connection"].sudo(f"sudo kubectl annotate nodes --all \
                                     --overwrite kubeadm.alpha.kubernetes.io/cri-socket=/run/containerd/containerd.sock"
                                     , is_async=False, hide=True)
        upgrade_group = kubernetes.get_group_for_upgrade(cluster)
        upgrade_group.sudo("rm -rf /var/run/docker.sock")

def kubernetes_apply_taints(cluster):
    # Apply taints after upgrade
    group = cluster.nodes['control-plane']
    kubernetes.apply_taints(group)

if __name__ == '__main__':
    main()



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io

from kubemarine.core import utils
from kubemarine.core.group import NodeGroup


def enrich_inventory(inventory, _):
    if not inventory["plugins"]["nginx-ingress-controller"]["install"]:
        return inventory

    if inventory["plugins"]["nginx-ingress-controller"].get('custom_headers'):
        if not inventory["plugins"]["nginx-ingress-controller"].get('config_map'):
            inventory["plugins"]["nginx-ingress-controller"]['config_map'] = {}
        if not inventory["plugins"]["nginx-ingress-controller"]['config_map'].get('proxy-set-headers'):
            inventory["plugins"]["nginx-ingress-controller"]['config_map']['proxy-set-headers'] = 'ingress-nginx/custom-headers'

    return inventory


def cert_renew_enrichment(inventory, cluster):
    # check that renewal is required for nginx
    if cluster.context.get('initial_procedure') != 'cert_renew' \
            or not cluster.procedure_inventory.get("nginx-ingress-controller"):
        return inventory

    nginx_plugin = inventory["plugins"]["nginx-ingress-controller"]

    # check that renewal is possible
    if not nginx_plugin["install"]:
        raise Exception("Certificates can not be renewed for nginx plugin since it is not installed")

    # update certificates in inventory
    nginx_plugin["controller"]["ssl"]["default-certificate"] = cluster.procedure_inventory["nginx-ingress-controller"]

    return inventory


def finalize_inventory(cluster, inventory_to_finalize):
    # check that renewal is required for nginx
    if cluster.context.get('initial_procedure') != 'cert_renew' \
            or not cluster.procedure_inventory.get("nginx-ingress-controller"):
        return inventory_to_finalize

    if not inventory_to_finalize["plugins"].get("nginx-ingress-controller"):
        inventory_to_finalize["plugins"]["nginx-ingress-controller"] = {}

    if not inventory_to_finalize["plugins"]["nginx-ingress-controller"].get("controller"):
        inventory_to_finalize["plugins"]["nginx-ingress-controller"]["controller"] = {}

    if not inventory_to_finalize["plugins"]["nginx-ingress-controller"]["controller"].get("ssl"):
        inventory_to_finalize["plugins"]["nginx-ingress-controller"]["controller"]["ssl"] = {}

    nginx_plugin = inventory_to_finalize["plugins"]["nginx-ingress-controller"]
    nginx_plugin["controller"]["ssl"]["default-certificate"] = cluster.procedure_inventory["nginx-ingress-controller"]

    return inventory_to_finalize


def manage_custom_certificate(cluster):
    if not cluster.inventory["plugins"]["nginx-ingress-controller"]["controller"]["ssl"].get("default-certificate"):
        cluster.log.debug("No custom default ingress certificate specified, skipping...")
        return

    base_path = "/etc/kubernetes/custom-certs"
    certificate_path = base_path + "/cert"
    private_key_path = base_path + "/key"
    secret_name = "default-ingress-cert"
    secret_namespace = "kube-system"

    first_control_plane = cluster.nodes["control-plane"].get_first_member()
    default_cert = cluster.inventory["plugins"]["nginx-ingress-controller"]["controller"]["ssl"]["default-certificate"]

    # first, we need to load cert and key files to first control-plane to known locations
    first_control_plane.sudo(f"mkdir -p {base_path}")
    try:
        first_control_plane.call(put_custom_certificate,
                          default_cert=default_cert,
                          crt_path=certificate_path,
                          key_path=private_key_path)

        # second, we need to validate cert and key using openssl
        first_control_plane.call(verify_certificate_and_key, crt_path=certificate_path, key_path=private_key_path)

        # third, we need to create tls secret under well-known name
        # this certificate is already configured to be used by controller
        first_control_plane.call(create_tls_secret,
                          crt_path=certificate_path,
                          key_path=private_key_path,
                          name=secret_name,
                          namespace=secret_namespace)
    finally:
        # fourth, we need to remove base path dir
        first_control_plane.sudo(f"rm -rf {base_path}")


def put_custom_certificate(first_control_plane: NodeGroup, default_cert, crt_path, key_path):
    if default_cert.get("data"):
        cert = io.StringIO(default_cert["data"]["cert"])
        key = io.StringIO(default_cert["data"]["key"])
    else:
        cert = io.StringIO(utils.read_external(default_cert["paths"]["cert"]))
        key = io.StringIO(utils.read_external(default_cert["paths"]["key"]))

    first_control_plane.put(cert, crt_path, sudo=True)
    first_control_plane.put(key, key_path, sudo=True)


def verify_certificate_and_key(first_control_plane: NodeGroup, crt_path, key_path):
    crt_md5 = first_control_plane.sudo(f"openssl x509 -noout -modulus -in {crt_path} | openssl md5").get_simple_out()
    key_md5 = first_control_plane.sudo(f"openssl rsa -noout -modulus -in {key_path} | openssl md5").get_simple_out()
    if crt_md5 != key_md5:
        raise Exception("Custom default ingress certificate and key are not compatible!")


def create_tls_secret(first_control_plane, crt_path, key_path, name, namespace):
    first_control_plane.sudo(f"kubectl create secret tls {name} --key {key_path} --cert {crt_path} -n {namespace} "
                      f"--dry-run -o yaml | sudo kubectl apply -f -", timeout=300)



#!/usr/bin/env python3
# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import OrderedDict

from kubemarine.core import flow
from kubemarine.core.action import Action
from kubemarine.core.resources import DynamicResources
from kubemarine.procedures import install
from kubemarine import system


def reboot(cluster):
    if cluster.context.get('initial_procedure') != 'reboot':
        raise ImportError('Invalid reboot.py usage, please use system.reboot_nodes')

    if not cluster.procedure_inventory.get("nodes"):
        cluster.log.verbose('No nodes defined in procedure: all nodes will be rebooted')
    else:
        cluster.log.verbose('There are nodes defined in procedure: only defined will be rebooted')

    nodes = []

    cluster.log.verbose('The following nodes will be rebooted:')
    for node in cluster.procedure_inventory.get("nodes", cluster.nodes['all'].get_ordered_members_list(provide_node_configs=True)):
        nodes.append(node['name'])
        cluster.log.verbose('  - ' + node['name'])

    system.reboot_group(cluster.make_group_from_nodes(nodes),
                        try_graceful=cluster.procedure_inventory.get("graceful_reboot"))


tasks = OrderedDict({
    "reboot": reboot,
    "overview": install.overview,
})


class RebootAction(Action):
    def __init__(self):
        super().__init__('reboot')

    def run(self, res: DynamicResources):
        flow.run_tasks(res, tasks)


def main(cli_arguments=None):
    cli_help = '''
    Script for Kubernetes nodes graceful rebooting.

    How to use:

    '''

    parser = flow.new_procedure_parser(cli_help, optional_config=True, tasks=tasks)

    context = flow.create_context(parser, cli_arguments, procedure='reboot')

    flow.run_actions(context, [RebootAction()])


if __name__ == '__main__':
    main()



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import textwrap
from traceback import *
import csv
from datetime import datetime
from kubemarine.core import utils
import fabric

TC_UNKNOWN = -1
TC_PASSED = 0
TC_FAILED = 1
TC_WARNED = 2
TC_EXCEPTED = 3

badges_weights = {
    'succeeded': 0,
    'warned': 1,
    'failed': 2,
    'excepted': 3,
    'unknown': 4,
}


class TestCase:

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if type is None:
            if self.status is TC_UNKNOWN:
                self.success()
        elif type is TestFailure:
            self.fail(value)
        elif type is TestWarn:
            self.warn(value)
        else:
            if isinstance(value, fabric.group.GroupException):
                value.result.print()
            else:
                print_exc()
            self.exception(value)
        print(self.get_summary(show_hint=True))
        return True

    def __init__(self, ts, id, category, name, default_results=None, minimal=None, recommended=None):
        self.include_in_ts(ts)
        self.category = category
        self.id = str(id)
        self.name = name
        self.status = TC_UNKNOWN
        self.results = default_results
        self.minimal = minimal
        self.recommended = recommended

    def include_in_ts(self, ts):
        ts.register_tc(self)
        return self

    def success(self, results=None):
        if self.results is None:
            self.results = results
        self.status = TC_PASSED
        return self

    def fail(self, results):
        self.status = TC_FAILED
        self.results = results
        return self

    def warn(self, results):
        self.status = TC_WARNED
        self.results = results
        return self

    def exception(self, results):
        self.status = TC_EXCEPTED
        self.results = results
        return self

    def get_summary(self, show_description=False, show_hint=False, show_minimal=False, show_recommended=False):
        output = ""

        output += " " * (15 - len(self.category))
        output += self.category + "  "

        color = ""
        if self.is_succeeded():
            color = "\x1b[38;5;041m"
            output += " \x1b[48;5;041m\x1b[38;5;232m   OK   \x1b[49m\x1b[39m  "
        if self.is_failed():
            color = "\x1b[38;5;196m"
            output += " \x1b[48;5;196m\x1b[38;5;231m  FAIL  \x1b[49m\x1b[39m  "
        if self.is_warned():
            color = "\x1b[38;5;208m"
            output += " \x1b[48;5;208m\x1b[38;5;231m  WARN  \x1b[49m\x1b[39m  "
        if self.is_excepted():
            color = "\x1b[31m"
            output += " \x1b[41m ERROR? \x1b[49m  "

        output += self.id + "  "
        output += self.name + " "

        results = " " + str(self.results)

        output += "." * (146 - len(output) - len(results))
        output += "%s%s\x1b[39m" % (color, results)

        if show_minimal:
            if self.minimal is None:
                output += ' ' * 15
            else:
                minimal = str(self.minimal)
                output += ' ' * (15-len(minimal)) + minimal

        if show_recommended:
            if self.recommended is None:
                output += ' ' * 14
            else:
                recommended = str(self.recommended)
                output += ' ' * (14-len(recommended)) + recommended

        if show_hint and (isinstance(self.results, TestFailure) or isinstance(self.results, TestWarn)) and self.results.hint is not None:
            output += "\n                  HINT:\n" + textwrap.indent(str(self.results.hint), "                       ")

        return output

    def get_readable_status(self):
        if self.is_succeeded():
            return 'ok'
        if self.is_failed():
            return 'fail'
        if self.is_warned():
            return 'warning'
        if self.is_excepted():
            return 'exception'

    def is_succeeded(self):
        return self.status is TC_PASSED

    def is_failed(self):
        return self.status is TC_FAILED

    def is_warned(self):
        return self.status is TC_WARNED

    def is_excepted(self):
        return self.status is TC_EXCEPTED


class TestCaseNegativeResult(BaseException):

    def __init__(self, message, hint=None, group_result=None):
        super().__init__(message)
        self.message = message
        self.hint = hint
        self.group_result = group_result


class TestFailure(TestCaseNegativeResult):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class TestWarn(TestCaseNegativeResult):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class TestSuite:

    def __init__(self):
        self.tcs = []

    def register_tc(self, tc):
        self.tcs.append(tc)

    def is_any_test_failed(self):
        for tc in self.tcs:
            if tc.is_failed() or tc.is_excepted():
                return True
        return False

    def is_any_test_warned(self):
        for tc in self.tcs:
            if tc.is_warned():
                return True
        return False

    def get_final_summary(self, show_minimal=True, show_recommended=True):
        result = "          Group    Status   ID    Test                                                               Actual result"
        if show_minimal:
            result += "        Minimal"
        if show_recommended:
            result += "   Recommended"
        result += "\n"

        for tc in self.tcs:
            result += "\n" + tc.get_summary(show_minimal=show_minimal, show_recommended=show_recommended)

        result += "\n\nOVERALL RESULTS: "

        for key, value in sorted(self.get_stats_data().items(), key=lambda _key: badges_weights[_key[0]]):
            colors = ''
            if key == 'succeeded':
                colors = "\x1b[48;5;041m\x1b[38;5;232m"
            if key == 'failed':
                colors = "\x1b[48;5;196m\x1b[38;5;231m"
            if key == 'warned':
                colors = "\x1b[48;5;208m\x1b[38;5;231m"
            if key == 'excepted':
                colors = "\x1b[41m"
            result += "%s %s %s \x1b[49m\x1b[39m " % (colors, value ,key.upper())

        result += "\n"

        return result

    def print_final_status(self, log):
        if self.is_any_test_failed():
            log.error("\nTEST FAILED"
                      "\nThe environment does not meet the minimal requirements. Check the test report and resolve the issues.")
            return
        if self.is_any_test_warned():
            log.warning("\nTEST PASSED WITH WARNINGS"
                        "\nThe environment meets the minimal requirements, but is not as recommended. Try to check the test report and resolve the issues.")
            return
        log.info("\nTEST PASSED")

    def get_stats_data(self):
        results = {}
        for tc in self.tcs:
            key = 'unknown'
            if tc.is_succeeded():
                key = 'succeeded'
            elif tc.is_failed():
                key = 'failed'
            elif tc.is_warned():
                key = 'warned'
            elif tc.is_excepted():
                key = 'excepted'
            value = results.get(key, 0) + 1
            results[key] = value
        return results

    def save_csv(self, destination_file_path, delimiter=';'):
        with utils.open_external(destination_file_path, mode='w') as stream:
            csv_writer = csv.writer(stream, delimiter=delimiter, quotechar='"', quoting=csv.QUOTE_MINIMAL)
            csv_writer.writerow(['group', 'status', 'test_id', 'test_name', 'current_result', 'minimal_result', 'recommended_result'])
            for tc in self.tcs:
                csv_writer.writerow([
                    tc.category.lower(),
                    tc.get_readable_status(),
                    tc.id,
                    tc.name,
                    tc.results,
                    tc.minimal,
                    tc.recommended
                ])

    def save_html(self, destination_file_path, check_type, append_styles=True):
        with utils.open_external(destination_file_path, mode='w') as stream:
            stream.write('<!DOCTYPE html><html><head><meta charset="utf-8"><title>%s Check Report</title></head><body><div id="date">%s</div><div id="stats">' % (check_type, datetime.utcnow()))
            for key, value in sorted(self.get_stats_data().items(), key=lambda _key: badges_weights[_key[0]]):
                stream.write('<div class="%s">%s %s</div>' % (key, value, key))
            stream.write('</div><h1>%s Check Report</h1><table>' % check_type)
            stream.write('<thead><tr><td>Group</td><td>Status</td><td>ID</td><td>Test</td><td>Actual Result</td><td>Minimal</td><td>Recommended</td></tr></thead><tbody>')
            for tc in self.tcs:
                minimal = tc.minimal
                if minimal is None:
                    minimal = ''
                recommended = tc.recommended
                if recommended is None:
                    recommended = ''
                stream.write('<tr class="%s"><td>%s</td><td><div>%s</div></td><td>%s</td><td>%s</td><td>%s</td><td>%s</td><td>%s</td></tr>' %
                             (tc.get_readable_status(),
                              tc.category.lower(),
                              tc.get_readable_status(),
                              tc.id,
                              tc.name,
                              tc.results,
                              minimal,
                              recommended
                              ))
            stream.write('</tbody></table>')
            if append_styles:
                css = utils.read_internal('resources/reports/check_report.css')
                stream.write('<style>\n%s\n</style>' % css)

            stream.write('</body></html>')



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import configparser
import io
import paramiko
import re
import socket
import time
from copy import deepcopy
from typing import Dict

from dateutil.parser import parse
import fabric
import yaml
from ordered_set import OrderedSet

from kubemarine import selinux, kubernetes, apparmor
from kubemarine.core import utils, static
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.executor import RemoteExecutor
from kubemarine.core.group import NodeGroupResult, NodeGroup
from kubemarine.core.yaml_merger import default_merger
from kubemarine.core.annotations import restrict_empty_group


def verify_inventory(inventory, cluster):

    if cluster.inventory['services']['ntp'].get('chrony', {}).get('servers') \
        and (cluster.inventory['services']['ntp'].get('timesyncd', {}).get('Time', {}).get('NTP') or
             cluster.inventory['services']['ntp'].get('timesyncd', {}).get('Time', {}).get('FallbackNTP')):
        raise Exception('chrony and timesyncd configured both at the same time')

    return inventory


def enrich_etc_hosts(inventory, cluster):
    control_plain = inventory['control_plain']['internal']

    control_plain_names = inventory['services']['etc_hosts'].get(control_plain, [])
    control_plain_names.append(cluster.inventory['cluster_name'])
    control_plain_names.append('control-plain')
    control_plain_names = list(OrderedSet(control_plain_names))
    inventory['services']['etc_hosts'][control_plain] = control_plain_names

    for node in cluster.inventory['nodes']:
        if 'remove_node' in node['roles']:
            continue

        internal_node_ip_names = inventory['services']['etc_hosts'].get(node['internal_address'], [])
        internal_node_ip_names.append("%s.%s" % (node['name'], cluster.inventory['cluster_name']))
        internal_node_ip_names.append(node['name'])
        internal_node_ip_names = list(OrderedSet(internal_node_ip_names))
        inventory['services']['etc_hosts'][node['internal_address']] = internal_node_ip_names

        if node.get('address'):
            external_node_ip_names = inventory['services']['etc_hosts'].get(node['address'], [])
            external_node_ip_names.append("%s-external.%s" % (node['name'], cluster.inventory['cluster_name']))
            external_node_ip_names.append(node['name'] + "-external")
            external_node_ip_names = list(OrderedSet(external_node_ip_names))
            inventory['services']['etc_hosts'][node['address']] = external_node_ip_names

    return inventory


def enrich_upgrade_inventory(inventory: dict, cluster: KubernetesCluster):
    if cluster.context.get("initial_procedure") != "upgrade":
        return inventory

    os_family = cluster.get_os_family()
    if os_family in ('unknown', 'unsupported', 'multiple'):
        raise Exception("Upgrade is possible only for cluster "
                        "with all nodes having the same and supported OS family")

    # validate all packages sections in procedure inventory
    base_associations = static.DEFAULTS["services"]["packages"]["associations"][os_family]

    cluster_associations = deepcopy(inventory["services"]["packages"]["associations"][os_family])
    previous_ver = cluster.context["initial_kubernetes_version"]
    upgrade_plan = cluster.procedure_inventory.get('upgrade_plan')
    for version in upgrade_plan:
        upgrade_associations = cluster.procedure_inventory.get(version, {}).get("packages", {}).get("associations", {})
        for package in get_system_packages(cluster):
            if base_associations[package]["package_name"] != cluster_associations[package]["package_name"] \
                    and not upgrade_associations.get(package, {}).get("package_name"):
                raise Exception(f"Associations are redefined for {package} in cluster.yaml for version {previous_ver}, "
                                f"but not present in procedure inventory for version {version}. "
                                f"Please, specify required associations explicitly in procedure inventory "
                                f"for all versions since {previous_ver}.")
            if upgrade_associations.get(package, {}).get("package_name"):
                cluster_associations[package]["package_name"] = upgrade_associations[package]["package_name"]
        previous_ver = version

    upgrade_required = get_system_packages_for_upgrade(cluster)
    cluster.context["packages"] = {"upgrade_required": upgrade_required}

    upgrade_ver = cluster.context["upgrade_version"]
    packages_section = deepcopy(cluster.procedure_inventory.get(upgrade_ver, {}).get("packages", {}))
    # Move associations to the OS family specific section, and then merge with associations from procedure.
    # This effectively allows to specify only global section but not for specific OS family.
    # This restriction is because system.enrich_upgrade_inventory goes after packages.enrich_inventory_associations,
    # but in future the restriction can be eliminated.
    associations = packages_section.pop("associations", {})
    default_merger.merge(inventory["services"]["packages"]["associations"][os_family], associations)

    for _type in ['install', 'upgrade', 'remove']:
        packages = packages_section.pop(_type, None)
        if packages is None:
            continue
        if isinstance(packages, list):
            packages = {'include': packages}
        default_merger.merge(inventory["services"]["packages"].setdefault(_type, {}), packages)

    # merge remained packages section
    default_merger.merge(inventory["services"]["packages"], packages_section)

    return inventory


def get_system_packages_for_upgrade(cluster):
    upgrade_ver = cluster.context["upgrade_version"]
    previous_ver = cluster.context["initial_kubernetes_version"]
    compatibility = cluster.globals["compatibility_map"]["software"]

    # handle special cases in which upgrade is not required for particular package
    cluster_associations = cluster.inventory["services"]["packages"]["associations"][cluster.get_os_family()]
    upgrade_associations = cluster.procedure_inventory.get(upgrade_ver, {}).get("packages", {}).get("associations", {})
    system_packages = get_system_packages(cluster)
    upgrade_required = list(system_packages)
    for package in system_packages:
        defined_association = upgrade_associations.get(package, {}).get("package_name")
        if defined_association and defined_association == cluster_associations[package]['package_name']:
            # case 1: package_name is defined in upgrade inventory but is equal to one already defined in cluster.yaml
            upgrade_required.remove(package)
        elif compatibility.get(package) and compatibility[package][upgrade_ver] == compatibility[package][previous_ver] \
                and not defined_association:
            # case 2: package_name is not defined in upgrade inventory and default versions are equal
            upgrade_required.remove(package)

    # all other packages should be updated
    return upgrade_required


def get_system_packages(cluster):
    return [cluster.inventory['services']['cri']['containerRuntime']]


def fetch_os_versions(cluster: KubernetesCluster):
    group = cluster.nodes['all'].get_accessible_nodes()
    '''
    For Red Hat, CentOS, Oracle Linux, and Ubuntu information in /etc/os-release /etc/redhat-release is sufficient but,
    Debian stores the full version in a special file. sed transforms version string, eg 10.10 becomes DEBIAN_VERSION="10.10"  
    '''

    return group.run(
        "cat /etc/*elease; cat /etc/debian_version 2> /dev/null | sed 's/\\(.\\+\\)/DEBIAN_VERSION=\"\\1\"/' || true")


def detect_os_family(cluster):
    results = fetch_os_versions(cluster)

    for connection, result in results.items():
        stdout = result.stdout.lower()

        version = None
        lines = ''

        version_regex = re.compile("\\s\\d*\\.\\d*", re.M)
        for line in stdout.split("\n"):
            if 'centos' in line or 'rhel' in line:
                # CentOS and Red Hat have a major version in VERSION_ID string
                matches = re.findall(version_regex, line)
                if matches:
                    version = matches[0].strip()
            if '=' in line:
                lines += line + "\n"

        os_release = configparser.ConfigParser()
        os_release.read_string("[system]\n" + lines)
        name = os_release.get("system", "id").replace('"', '')
        if version is None:
            if name == 'debian':
                version = os_release.get("system", "debian_version").replace('"', '')
            else:
                # Oracle Linux and Ubuntu have full version in VERSION_ID string
                version = os_release.get("system", "version_id").replace('"', '')

        cluster.log.debug("Distribution: %s; Version: %s" % (name, version))

        os_family = 'unsupported'
        if name in cluster.globals["compatibility_map"]["distributives"]:
            os_family = 'unknown'
            os_family_list = cluster.globals["compatibility_map"]["distributives"][name]
            for os_family_item in os_family_list:
                if version in os_family_item["versions"]:
                    os_family = os_family_item["os_family"]
                    break

        cluster.log.debug("OS family: %s" % os_family)

        cluster.context["nodes"][connection.host]["os"] = {
            'name': name,
            'version': version,
            'family': os_family
        }


def get_compatibility_version_key(cluster: KubernetesCluster) -> str or None:
    """
    Get os-specific version key to be used in software compatibility map.
    :param cluster: Cluster object for which to resolve compatibility version key.
    :return: String to use as version key. None if OS is unknown or multiple OS present.
    """
    """
    Return os-specific version compatibility key.
    If OS is unknown or multiple OS, then returns None.
    """
    os = cluster.get_os_family()
    if os == "rhel":
        return "version_rhel"
    elif os == "rhel8":
        return "version_rhel8"
    elif os == "debian":
        return "version_debian"
    else:
        return None


def update_resolv_conf(group, config=None):
    if config is None:
        raise Exception("Data can't be empty")

    # TODO: Use Jinja template
    buffer = get_resolv_conf_buffer(config)
    utils.dump_file(group.cluster, buffer, 'resolv.conf')
    group.put(buffer, "/etc/resolv.conf", backup=True, immutable=True, sudo=True, hide=True)


def get_resolv_conf_buffer(config):
    buffer = io.StringIO()
    if config.get("search") is not None:
        buffer.write("search %s\n" % config["search"])
    if config.get("nameservers") is not None:
        for address in config.get("nameservers"):
            buffer.write("nameserver %s\n" % address)
    return buffer


def generate_etc_hosts_config(inventory, cluster=None):
    result = ""

    max_len_ip = 0

    ignore_ips = []
    if cluster and cluster.context['initial_procedure'] == 'remove_node':
        for removal_node in cluster.procedure_inventory.get("nodes"):
            removal_node_name = removal_node['name']
            for node in inventory['nodes']:
                if node['name'] == removal_node_name:
                    if node.get('address'):
                        ignore_ips.append(node['address'])
                    if node.get('internal_address'):
                        ignore_ips.append(node['internal_address'])

    ignore_ips = list(set(ignore_ips))

    for ip in list(inventory['services']['etc_hosts'].keys()):
        if len(ip) > max_len_ip:
            max_len_ip = len(ip)

    for ip, names in inventory['services']['etc_hosts'].items():
        if isinstance(names, list):
            # remove records with empty values from list
            names = list(filter(len, names))
            # if list is empty, then skip
            if not names:
                continue
            names = " ".join(names)
        if ip not in ignore_ips:
            result += "%s%s  %s\n" % (ip, " " * (max_len_ip - len(ip)), names)

    return result


def update_etc_hosts(group, config=None):
    if config is None:
        raise Exception("Data can't be empty")
    utils.dump_file(group.cluster, config, 'etc_hosts')
    group.put(io.StringIO(config), "/etc/hosts", backup=True, sudo=True, hide=True)


def stop_service(group: NodeGroup, name: str) -> NodeGroupResult:
    return group.sudo('systemctl stop %s' % name)


def start_service(group: NodeGroup, name: str) -> NodeGroupResult:
    return group.sudo('systemctl start %s' % name)


def restart_service(group, name=None):
    if name is None:
        raise Exception("Service name can't be empty")
    return group.sudo('systemctl restart %s' % name)


def enable_service(group, name=None, now=True):
    if name is None:
        raise Exception("Service name can't be empty")

    cmd = 'systemctl enable %s' % name
    if now:
        cmd = cmd + " --now"
    return group.sudo(cmd)


def disable_service(group, name=None, now=True):
    if name is None:
        raise Exception("Service name can't be empty")

    cmd = 'systemctl disable %s' % name
    if now:
        cmd = cmd + " --now"
    return group.sudo(cmd)


def patch_systemd_service(group: NodeGroup, service_name: str, patch_source: str):
    group.sudo(f"mkdir -p /etc/systemd/system/{service_name}.service.d")
    group.put(io.StringIO(utils.read_internal(patch_source)),
              f"/etc/systemd/system/{service_name}.service.d/{service_name}.conf",
              sudo=True)
    group.sudo("systemctl daemon-reload")


def fetch_firewalld_status(group: NodeGroup) -> NodeGroupResult:
    return group.sudo("systemctl status firewalld", warn=True)


def is_firewalld_disabled(group):
    result = fetch_firewalld_status(group)
    disabled_status = True

    for node_result in list(result.values()):
        if node_result.return_code != 4 and "disabled" not in node_result.stdout:
            disabled_status = False

    return disabled_status, result


def disable_firewalld(group):
    log = group.cluster.log

    already_disabled, result = is_firewalld_disabled(group)

    if already_disabled:
        log.debug("Skipped - FirewallD already disabled or not installed")
        return result

    log.verbose("Trying to stop and disable FirewallD...")

    result = disable_service(group, name='firewalld', now=True)

    group.cluster.schedule_cumulative_point(reboot_nodes)
    group.cluster.schedule_cumulative_point(verify_system)

    return result


def is_swap_disabled(group):
    result = group.sudo("cat /proc/swaps", warn=True)
    disabled_status = True

    for node_result in list(result.values()):
        # is there any other lines excluding first head line?
        if node_result.stdout.strip().split('\n')[1:]:
            disabled_status = False

    return disabled_status, result


def disable_swap(group):
    log = group.cluster.log

    already_disabled, result = is_swap_disabled(group)

    if already_disabled:
        log.debug("Skipped - swap already disabled")
        return result

    log.verbose("Switching swap off...")

    group.sudo('swapoff -a', warn=True)
    group.sudo('sed -i.bak \'/swap/d\' /etc/fstab', warn=True)

    group.cluster.schedule_cumulative_point(reboot_nodes)
    group.cluster.schedule_cumulative_point(verify_system)


def reboot_nodes(cluster: KubernetesCluster):
    cluster.nodes["all"].get_new_nodes_or_self().call(reboot_group)


def reboot_group(group: NodeGroup, try_graceful=None):
    log = group.cluster.log

    if try_graceful is None:
        if 'controlplain_uri' not in group.cluster.context.keys():
            kubernetes.is_cluster_installed(group.cluster)

    graceful_reboot = try_graceful is True or \
                      (try_graceful is None and group.cluster.context['controlplain_uri'] is not None)

    if not graceful_reboot:
        return perform_group_reboot(group)

    log.verbose('Graceful reboot required')

    first_control_plane = group.cluster.nodes['control-plane'].get_first_member()
    results = NodeGroupResult(group.cluster)

    for node in group.get_ordered_members_list(provide_node_configs=True):
        cordon_required = 'control-plane' in node['roles'] or 'worker' in node['roles']
        if cordon_required:
            res = first_control_plane.sudo(
                kubernetes.prepare_drain_command(node, group.cluster.inventory['services']['kubeadm']['kubernetesVersion'],
                                                 group.cluster.globals, False, group.cluster.nodes), warn=True)
            log.verbose(res)
        log.debug(f'Rebooting node "{node["name"]}"')
        raw_results = perform_group_reboot(node['connection'])
        if cordon_required:
            res = first_control_plane.sudo(f'kubectl uncordon {node["name"]}', warn=True)
            log.verbose(res)
        results.update(raw_results)

    return results


def get_reboot_history(group: NodeGroup):
    return group.sudo('last reboot')


def perform_group_reboot(group: NodeGroup):
    log = group.cluster.log

    initial_boot_history = get_reboot_history(group)
    result = group.sudo(group.cluster.globals['nodes']['boot']['reboot_command'], warn=True)
    log.debug("Waiting for boot up...")
    group.wait_for_reboot(initial_boot_history)
    return result


def reload_systemctl(group):
    return group.sudo('systemctl daemon-reload')


def add_to_path(group, string):
    # TODO: Also update PATH in ~/.bash_profile
    group.sudo("export PATH=$PATH:%s" % string)

def configure_chronyd(group, retries=60):
    cluster = group.cluster
    log = cluster.log
    chronyd_config = ''

    for server in cluster.inventory['services']['ntp']['chrony']['servers']:
        chronyd_config += "server " + server + "\n"

    if cluster.inventory['services']['ntp']['chrony'].get('makestep'):
        chronyd_config += "\nmakestep " + cluster.inventory['services']['ntp']['chrony']['makestep']

    if cluster.inventory['services']['ntp']['chrony'].get('rtcsync', False):
        chronyd_config += "\nrtcsync"

    utils.dump_file(cluster, chronyd_config, 'chrony.conf')
    group.put(io.StringIO(chronyd_config), '/etc/chrony.conf', backup=True, sudo=True)
    group.sudo('systemctl restart chronyd')
    while retries > 0:
        log.debug("Waiting for time sync, retries left: %s" % retries)
        results = group.sudo('chronyc tracking && sudo chronyc sources')
        if results.stdout_contains("Normal"):
            log.verbose("NTP service reported successful time synchronization, validating...")

            current_node_time, nodes_time, time_diff = get_nodes_time(group)
            if time_diff > cluster.globals['nodes']['max_time_difference']:
                log.debug("Time is not synced yet")
                log.debug(results)
            else:
                log.debug("Time synced!")
                return results

        else:
            log.debug("Time is not synced yet")
            log.debug(results)
        time.sleep(1)
        retries -= 1

    raise Exception("Time not synced, but timeout is reached")


def configure_timesyncd(group, retries=120):
    cluster = group.cluster
    log = cluster.log
    timesyncd_config = ''

    for section, options in cluster.inventory['services']['ntp']['timesyncd'].items():
        timesyncd_config += '[%s]' % section
        for option_name, option_value in options.items():
            if isinstance(option_value, list):
                option_value_str = " ".join(option_value)
            else:
                option_value_str = str(option_value)
            timesyncd_config += '\n%s=%s' % (option_name, option_value_str)
        timesyncd_config += '\n\n'

    utils.dump_file(cluster, timesyncd_config, 'timesyncd.conf')
    group.put(io.StringIO(timesyncd_config), '/etc/systemd/timesyncd.conf', backup=True, sudo=True)
    res = group.sudo('timedatectl set-ntp true '
                     '&& sudo systemctl enable --now systemd-timesyncd.service '
                     '&& sudo systemctl restart systemd-timesyncd.service '
                     '&& sudo systemctl status systemd-timesyncd.service')
    log.verbose(res)
    while retries > 0:
        log.debug("Waiting for time sync, retries left: %s" % retries)
        results = group.sudo('timedatectl timesync-status && sudo timedatectl status')
        if results.stdout_contains("synchronized: yes"):
            log.verbose("NTP service reported successful time synchronization, validating...")

            current_node_time, nodes_time, time_diff = get_nodes_time(group)
            if time_diff > cluster.globals['nodes']['max_time_difference']:
                log.debug("Time is not synced yet")
                log.debug(results)
            else:
                log.debug("Time synced!")
                return results

        else:
            log.debug("Time is not synced yet")
            log.debug(results)
        time.sleep(1)
        retries -= 1

    raise Exception("Time not synced, but timeout is reached")


def setup_modprobe(group):
    log = group.cluster.log

    if group.cluster.inventory['services'].get('modprobe') is None \
            or not group.cluster.inventory['services']['modprobe']:
        log.debug('Skipped - no modprobe configs in inventory')
        return

    is_valid, result = is_modprobe_valid(group)

    if is_valid:
        log.debug("Skipped - all necessary kernel modules are presented")
        return result

    config = ''
    raw_config = ''
    for module_name in group.cluster.inventory['services']['modprobe']:
        module_name = module_name.strip()
        if module_name is not None and module_name != '':
            config += module_name + "\n"
            raw_config += module_name + " "

    log.debug("Uploading config...")
    utils.dump_file(group.cluster, config, 'modprobe_predefined.conf')
    group.put(io.StringIO(config), "/etc/modules-load.d/predefined.conf", backup=True, sudo=True, hide=True)
    group.sudo("modprobe -a %s" % raw_config)

    group.cluster.schedule_cumulative_point(reboot_nodes)
    group.cluster.schedule_cumulative_point(verify_system)


def is_modprobe_valid(group):
    log = group.cluster.log

    verify_results = group.sudo("lsmod", warn=True)
    is_valid = True

    for module_name in group.cluster.inventory['services']['modprobe']:
        for conn, result in verify_results.items():
            if module_name not in result.stdout:
                log.debug('Kernel module %s not found at %s' % (module_name, conn.host))
                is_valid = False

    return is_valid, verify_results


def verify_system(cluster: KubernetesCluster):
    group = cluster.nodes["all"].get_new_nodes_or_self()
    log = cluster.log
    # this method handles clusters with multiple OS
    os_family = group.get_nodes_os()

    if os_family in ['rhel', 'rhel8'] and cluster.is_task_completed('prepare.system.setup_selinux'):
        log.debug("Verifying Selinux...")
        selinux_configured, selinux_result, selinux_parsed_result = \
            selinux.is_config_valid(group,
                                    state=selinux.get_expected_state(cluster.inventory),
                                    policy=selinux.get_expected_policy(cluster.inventory),
                                    permissive=selinux.get_expected_permissive(cluster.inventory))
        log.debug(selinux_result)
        if not selinux_configured:
            raise Exception("Selinux is still not configured")
    else:
        log.debug('Selinux verification skipped - origin task was not completed')

    if cluster.is_task_completed('prepare.system.setup_apparmor') and os_family == 'debian':
        log.debug("Verifying Apparmor...")
        expected_profiles = cluster.inventory['services']['kernel_security'].get('apparmor', {})
        apparmor_configured, result = apparmor.is_state_valid(group, expected_profiles)
        if not apparmor_configured:
            raise Exception("Apparmor is still not configured")
    else:
        log.debug('Apparmor verification skipped - origin task was not completed')

    if cluster.is_task_completed('prepare.system.disable_firewalld'):
        log.debug("Verifying FirewallD...")
        firewalld_disabled, firewalld_result = is_firewalld_disabled(group)
        log.debug(firewalld_result)
        if not firewalld_disabled:
            raise Exception("FirewallD is still enabled")
    else:
        log.debug('FirewallD verification skipped - origin disable task was not completed')

    if cluster.is_task_completed('prepare.system.disable_swap'):
        log.debug("Verifying swap...")
        swap_disabled, swap_result = is_swap_disabled(group)
        log.debug(swap_result)
        if not swap_disabled:
            raise Exception("Swap is still enabled")
    else:
        log.debug('Swap verification skipped - origin disable task was not completed')

    if cluster.is_task_completed('prepare.system.modprobe'):
        log.debug("Verifying modprobe...")
        modprobe_valid, modprobe_result = is_modprobe_valid(group)
        log.debug(modprobe_result)
        if not modprobe_valid:
            raise Exception("Required kernel modules are not presented")
    else:
        log.debug('Modprobe verification skipped - origin setup task was not completed')


def detect_active_interface(cluster: KubernetesCluster):
    group = cluster.nodes['all'].get_accessible_nodes()
    with RemoteExecutor(cluster) as exe:
        for node in group.get_ordered_members_list(provide_node_configs=True):
            detect_interface_by_address(node['connection'], node['internal_address'])
    for cxn, host_results in exe.get_last_results().items():
        try:
            interface = list(host_results.values())[0].stdout.strip()
        except Exception:
            interface = None
        cluster.context['nodes'][cxn.host]['active_interface'] = interface

    return exe.get_last_results_str()


def detect_interface_by_address(group: NodeGroup, address: str):
    return group.run("/usr/sbin/ip -o a | grep %s | awk '{print $2}'" % address)


def _detect_nodes_access_info(cluster: KubernetesCluster):
    nodes_context = cluster.context['nodes']
    hosts_unknown_status = [host for host, node_context in nodes_context.items() if 'access' not in node_context]
    group_unknown_status = cluster.make_group(hosts_unknown_status)
    if group_unknown_status.is_empty():
        return

    check_active_timeout = int(cluster.globals["nodes"]["remove"]["check_active_timeout"])
    exc = None
    try:
        # This should invoke sudo last reboot
        results = group_unknown_status.wait_and_get_boot_history(timeout=check_active_timeout)
    except fabric.group.GroupException as e:
        exc = e
        results = e.result

    for connection, result in results.items():
        access_info = {
            'online': False,
            'accessible': False,
            'sudo': 'No'
        }
        nodes_context[connection.host]['access'] = access_info

        if isinstance(result, Exception):
            if NodeGroup.is_require_nopasswd_exception(result):
                # The error is thrown only if connection is successful, but something is wrong with sudo access.
                # In general, sudo password is incorrect. In our case, user is not a sudoer, or not a nopasswd sudoer.
                access_info['online'] = True
                access_info['accessible'] = True
            elif isinstance(result, socket.timeout):
                # Usually when node is off. All statuses are unchecked.
                pass
            elif isinstance(result, paramiko.ssh_exception.SSHException):
                # At least, node is on, but something is wrong with ssh credentials (user / identity key)
                access_info['online'] = True
            elif isinstance(result, paramiko.ssh_exception.NoValidConnectionsError):
                # Internal socket error, for example, when ssh daemon is off. All statuses are unchecked.
                pass
            else:
                raise exc
        else:
            access_info['online'] = True
            access_info['accessible'] = True
            access_info['sudo'] = "Yes"


def whoami(cluster: KubernetesCluster) -> NodeGroupResult:
    '''
    Determines different nodes access information, such as if the node is online, ssh credentials are correct, etc.
    '''
    _detect_nodes_access_info(cluster)

    results = cluster.nodes["all"].get_sudo_nodes().sudo("whoami")
    for connection, result in results.items():
        node_ctx = cluster.context['nodes'][connection.host]
        node_ctx['access']['sudo'] = 'Root' if result.stdout.strip() == "root" else 'Yes'
    return results


@restrict_empty_group
def get_nodes_time(group: NodeGroup) -> (float, Dict[fabric.connection.Connection, float], float):
    """
    Polls the time from the specified group of nodes, parses it and returns tuple with results.
    :param group: Group of nodes, where timestamps should be detected.
    :return: tuple with max timestamp, dict of parsed timestamps per node and max found time difference.

    Max time required for comparing dates computed on nodes, because we can not detect real time between nodes.
    Max nodes time should be less than minimal compared time from future.

    Timestamp - is a time in milliseconds since epoch.
    """

    # Please, note: this method can not detect time on nodes precisely, since Kubemarine can execute commands not at the
    # same time depending on various factors, for example, restrictions on the number of open sockets.

    parsed_time_per_node: Dict[fabric.connection.Connection, float] = {}

    min_time = None
    max_time = None

    # TODO: request and parse more accurate timestamp in milliseconds

    raw_results = group.run('date')
    for host, result in raw_results.items():
        parsed_time = parse(result.stdout.strip()).timestamp() * 1000
        parsed_time_per_node[host] = parsed_time
        if min_time is None or min_time > parsed_time:
            min_time = parsed_time
        if max_time is None or max_time < parsed_time:
            max_time = parsed_time

    return max_time, parsed_time_per_node, max_time-min_time



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from kubemarine import system


def get_status(group):
    log = group.cluster.log
    result = group.sudo("apparmor_status --json")
    parsed_result = {}
    if result:
        for connection, node_result in result.items():
            log.verbose('Parsing status for %s...' % connection.host)
            parsed_result[connection] = parse_status(node_result.stdout)
    print_status(log, parsed_result)
    return parsed_result


def parse_status(result_stdout):
    result = {}
    parsed_data = json.loads(result_stdout)
    modes_set = set()
    for mode in parsed_data['profiles'].values():
        modes_set.add(mode)
    for mode in modes_set:
        profile_list = []
        for profile_name, profile_state in parsed_data['profiles'].items():
            if profile_state == mode:
                profile_list.append(profile_name)
        result[profile_state] = profile_list
    return result


def print_status(log, parsed_result):
    res = "AppArmor Status:"
    for state in parsed_result.keys():
        res += "\n  Profiles in %s mode:" % state
        for profile in parsed_result[state]:
            res += "\n    - %s" % profile
    log.verbose(res)


def is_state_valid(group, expected_profiles):
    log = group.cluster.log

    log.verbose('Verifying Apparmor modes...')

    parsed_result = get_status(group)
    valid = True

    for connection, status in parsed_result.items():
        for state, profiles in expected_profiles.items():
            if not profiles:
                continue
            if state == 'disable':
                for profile in profiles:
                    for remote_profiles in status.values():
                        if profile in remote_profiles:
                            valid = False
                            log.verbose('Mode %s is enabled on remote host %s' % (state, connection.host))
                            break
            else:
                if not status.get(state):
                    valid = False
                    log.verbose('Mode %s is not presented on remote host %s' % (state, connection.host))
                    break
                # check if all 'cluster.yaml' settings reflect on particular node
                for profile in profiles:
                    if profile not in status[state]:
                        valid = False
                        log.verbose('Profile %s is not enabled in %s mode on remote host %s' % (profile, state, connection.host))
                        break

    return valid, parsed_result

# TODO: describe what the purpose of that method is
def convert_profile(profile):
    profile = profile.replace('/', '.')
    if profile[0] == '.':
        profile = profile[1:]
    return profile


def configure_apparmor(group, expected_profiles):
    cmd = ''
    for profile in expected_profiles.get('enforce', []):
        profile = convert_profile(profile)
        cmd += 'sudo rm -f /etc/apparmor.d/disable/%s; sudo rm -f /etc/apparmor.d/force-complain/%s; ' % (profile, profile)
    for profile in expected_profiles.get('complain', []):
        profile = convert_profile(profile)
        cmd += 'sudo rm -f /etc/apparmor.d/disable/%s; sudo ln -s /etc/apparmor.d/%s /etc/apparmor.d/force-complain/; ' % (profile, profile)
    for profile in expected_profiles.get('disable', []):
        profile = convert_profile(profile)
        cmd += 'sudo rm -f /etc/apparmor.d/force-complain/%s; sudo ln -s /etc/apparmor.d/%s /etc/apparmor.d/disable/; ' % (profile, profile)
    cmd += 'sudo systemctl reload apparmor.service && sudo apparmor_status'
    return group.sudo(cmd)


def setup_apparmor(group):
    log = group.cluster.log

    if group.get_nodes_os() != 'debian':
        log.debug("Skipped - Apparmor is supported only on Ubuntu/Debian")
        return

    expected_profiles = group.cluster.inventory['services']['kernel_security'].get('apparmor', {})
    valid, parsed_result = is_state_valid(group, expected_profiles)

    if valid:
        log.debug("Skipped - Apparmor already correctly configured")
        return

    log.debug(configure_apparmor(group, expected_profiles))
    group.cluster.schedule_cumulative_point(system.reboot_nodes)
    group.cluster.schedule_cumulative_point(system.verify_system)



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import hashlib
import io
import os
import random
import re
import time
import uuid
from datetime import datetime
from typing import Callable, Dict, List, Union, IO

import fabric
import invoke
from invoke import UnexpectedExit

from kubemarine.core.connections import Connections
from kubemarine.core.executor import RemoteExecutor

_GenericResult = Union[Exception, fabric.runners.Result, fabric.transfer.Result]
_HostToResult = Dict[str, _GenericResult]


# fabric.runners.Result is not equitable OOB, let it make equitable
def _compare_fabric_results(self: fabric.runners.Result, other) -> bool:
    if not isinstance(other, fabric.runners.Result):
        return False

    # todo should other fields be compared? Or probably custom class should be used to store result.
    return self.exited == other.exited \
        and self.stdout == other.stdout \
        and self.stderr == other.stderr


fabric.runners.Result.__eq__ = _compare_fabric_results
fabric.runners.Result.__ne__ = lambda self, other: not _compare_fabric_results(self, other)


class NodeGroupResult(fabric.group.GroupResult, Dict[fabric.connection.Connection, _GenericResult]):

    def __init__(self, cluster, results: _HostToResult or NodeGroupResult = None) -> None:
        super().__init__()

        self.cluster = cluster

        if results is not None:
            for host, result in results.items():
                if isinstance(results, NodeGroupResult):
                    host = host.host
                connection = cluster.nodes['all'].nodes.get(host)
                if connection is None:
                    raise Exception(f'Host "{host}" was not found in provided cluster object')
                self[connection] = result

    def get_simple_out(self) -> str:
        if len(self) != 1:
            raise NotImplementedError("Simple output can be returned only for NodeGroupResult consisted of "
                                      "exactly one node, but %s were provided." % list(self.keys()))

        res = list(self.values())[0]
        if not isinstance(res, fabric.runners.Result):
            raise NotImplementedError("It does not make sense to return simple output for result of type %s"
                                      % type(res))

        return res.stdout

    def __str__(self) -> str:
        output = ""
        for conn, result in self.items():

            # TODO: support print other possible exceptions
            if isinstance(result, invoke.exceptions.UnexpectedExit):
                result = result.result

            # for now we do not know how-to print transfer result
            if not isinstance(result, fabric.runners.Result):
                continue

            if output != "":
                output += "\n"
            output += "\t%s: code=%i" % (conn.host, result.exited)
            if result.stdout:
                output += "\n\t\tSTDOUT: %s" % result.stdout.replace("\n", "\n\t\t        ")
            if result.stderr:
                output += "\n\t\tSTDERR: %s" % result.stderr.replace("\n", "\n\t\t        ")
        return output

    def print(self) -> None:
        """
        Prints debug message to log with results for each node
        :return: None
        """
        self.cluster.log.debug(self)

    def get_group(self) -> NodeGroup:
        """
        Forms and returns a new group from node results
        :return: NodeGroup
        """
        hosts = []
        for connection in list(self.keys()):
            hosts.append(connection.host)
        return self.cluster.make_group(hosts)

    def is_any_has_code(self, code: int or str) -> bool:
        """
        Returns true if some group result has an exit code equal to the given one. Exceptions and other objects in
        results will be ignored.
        :param code: The code with which the result codes will be compared
        :return: Boolean
        """
        for conn, result in self.items():
            if isinstance(result, fabric.runners.Result) and str(result.exited) == str(code):
                return True
        return False

    def is_any_excepted(self) -> bool:
        """
        Returns true if at least one result in group is an execution
        :return: Boolean
        """
        for conn, result in self.items():
            if isinstance(result, Exception):
                return True
        return False

    def is_any_failed(self) -> bool:
        """
        Returns true if at least one result in the group finished with code 1 or failed with an exception
        :return: Boolean
        """
        return self.is_any_has_code(1) or self.is_any_excepted()

    def get_excepted_nodes_list(self) -> List[fabric.connection.Connection]:
        """
        Returns a list of nodes connections, for which the result is an exception
        :return: List with nodes connections
        """
        failed_nodes: List[fabric.connection.Connection] = []
        for conn, result in self.items():
            if isinstance(result, Exception):
                failed_nodes.append(conn)
        return failed_nodes

    def get_excepted_nodes_group(self) -> NodeGroup:
        """
        Forms and returns new NodeGroup of nodes, for which the result is an exception
        :return: NodeGroup:
        """
        nodes_list = self.get_excepted_nodes_list()
        return self.cluster.make_group(nodes_list)

    def get_exited_nodes_list(self) -> List[fabric.connection.Connection]:
        """
        Returns a list of nodes connections, for which the result is the completion of the command and the formation of
        the Fabric Result
        :return: List with nodes connections
        """
        failed_nodes: List[fabric.connection.Connection] = []
        for conn, result in self.items():
            if isinstance(result, fabric.runners.Result):
                failed_nodes.append(conn)
        return failed_nodes

    def get_exited_nodes_group(self) -> NodeGroup:
        """
        Forms and returns new NodeGroup of nodes, for which the result is the completion of the command and the
        formation of the Fabric Result
        :return: NodeGroup:
        """
        nodes_list = self.get_exited_nodes_list()
        return self.cluster.make_group(nodes_list)

    def get_failed_nodes_list(self) -> List[fabric.connection.Connection]:
        """
        Returns a list of nodes connections that either exited with an exception, or the exit code is equals 1
        :return: List with nodes connections
        """
        failed_nodes: List[fabric.connection.Connection] = []
        for conn, result in self.items():
            if isinstance(result, Exception) or result.exited == 1:
                failed_nodes.append(conn)
        return failed_nodes

    def get_failed_nodes_group(self) -> NodeGroup:
        """
        Forms and returns new NodeGroup of nodes that either exited with an exception, or the exit code is equals 1
        :return: NodeGroup:
        """
        nodes_list = self.get_failed_nodes_list()
        return self.cluster.make_group(nodes_list)

    def get_nonzero_nodes_list(self) -> List[fabric.connection.Connection]:
        """
        Returns a list of nodes connections that exited with non-zero exit code
        :return: List with nodes connections
        """
        nonzero_nodes: List[fabric.connection.Connection] = []
        for conn, result in self.items():
            if isinstance(result, Exception) or result.exited != 0:
                nonzero_nodes.append(conn)
        return nonzero_nodes

    def get_nonzero_nodes_group(self) -> NodeGroup:
        """
        Forms and returns new NodeGroup of nodes that exited with non-zero exit code
        :return: NodeGroup:
        """
        nodes_list = self.get_nonzero_nodes_list()
        return self.cluster.make_group(nodes_list)

    def get_nodes_list_where_value_in_stdout(self, value: str) -> List[fabric.connection.Connection]:
        """
        Returns a list of node connections that contains the given string value in results stderr.
        :param value: The string value to be found in the nodes results stderr.
        :return: List with nodes connections
        """
        nodes_with_stderr_value: List[fabric.connection.Connection] = []
        for conn, result in self.items():
            if isinstance(result, fabric.runners.Result) and value in result.stdout:
                nodes_with_stderr_value.append(conn)
        return nodes_with_stderr_value

    def get_nodes_list_where_value_in_stderr(self, value: str) -> List[fabric.connection.Connection]:
        """
        Returns a list of node connections that contains the given string value in results stderr.
        :param value: The string value to be found in the nodes results stderr.
        :return: List with nodes connections
        """
        nodes_with_stderr_value: List[fabric.connection.Connection] = []
        for conn, result in self.items():
            if isinstance(result, fabric.runners.Result) and value in result.stderr:
                nodes_with_stderr_value.append(conn)
        return nodes_with_stderr_value

    def get_nodes_group_where_value_in_stdout(self, value: str) -> NodeGroup:
        """
        Forms and returns new NodeGroup of nodes that contains the given string value in results stdout.
        :param value: The string value to be found in the nodes results stdout.
        :return: NodeGroup
        """
        nodes_list = self.get_nodes_list_where_value_in_stdout(value)
        return self.cluster.make_group(nodes_list)

    def get_nodes_group_where_value_in_stderr(self, value: str) -> NodeGroup:
        """
        Forms and returns new NodeGroup of nodes that contains the given string value in results stderr.
        :param value: The string value to be found in the nodes results stderr.
        :return: NodeGroup
        """
        nodes_list = self.get_nodes_list_where_value_in_stderr(value)
        return self.cluster.make_group(nodes_list)

    def stdout_contains(self, value: str) -> bool:
        """
        Checks for the presence of the given string in all results stdout.
        :param value: The string value to be found in the nodes results stdout.
        :return: true if string presented
        """
        return len(self.get_nodes_list_where_value_in_stdout(value)) > 0

    def stderr_contains(self, value: str) -> bool:
        """
        Checks for the presence of the given string in all results stderr.
        :param value: The string value to be found in the nodes results stderr.
        :return: true if string presented
        """
        return len(self.get_nodes_list_where_value_in_stderr(value)) > 0

    def __eq__(self, other) -> bool:
        if self is other:
            return True

        if not isinstance(other, NodeGroupResult):
            return False

        if len(self) != len(other):
            return False

        for conn, result in self.items():
            compared_result = other.get(conn)
            if compared_result is None:
                return False

            if not isinstance(result, fabric.runners.Result) or not isinstance(compared_result, fabric.runners.Result):
                raise NotImplementedError('Currently only instances of fabric.runners.Result can be compared')

            if result != compared_result:
                return False

        return True

    def __ne__(self, other) -> bool:
        return not self == other


class NodeGroup:

    def __init__(self, connections: Connections, cluster):
        from kubemarine.core.cluster import KubernetesCluster

        self.cluster: KubernetesCluster = cluster
        self.nodes = connections

    def __eq__(self, other):
        if self is other:
            return True

        if not isinstance(other, NodeGroup):
            return False

        if self.cluster != other.cluster:
            return False

        if len(self.nodes.keys()) != len(other.nodes.keys()):
            return False

        for host, connection in self.nodes.items():
            other_host_conn = other.nodes.get(host)
            if other_host_conn is None:
                return False
            if other_host_conn != connection:
                return False

        return True

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        # TODO: include cluster object and real connections into hash value
        nodes_addresses = tuple(self.nodes.keys())
        return hash(nodes_addresses)

    def _make_result(self, results: _HostToResult) -> NodeGroupResult:
        group_result = NodeGroupResult(self.cluster, results)
        return group_result

    def _make_result_or_fail(self, results: _HostToResult,
                             failure_criteria: Callable[[str, _GenericResult], bool]) -> NodeGroupResult:
        failed_hosts = [host for host, result in results.items() if failure_criteria(host, result)]
        group_result = self._make_result(results)

        if failed_hosts:
            raise fabric.group.GroupException(group_result)

        return group_result

    def run(self, *args, **kwargs) -> Union[NodeGroupResult, int]:
        return self.do("run", *args, **kwargs)

    def sudo(self, *args, **kwargs) -> Union[NodeGroupResult, int]:
        return self.do("sudo", *args, **kwargs)

    def put(self, local_file: Union[io.StringIO, str], remote_file: str, **kwargs):
        if isinstance(local_file, io.StringIO):
            self.cluster.log.verbose("Text is being transferred to remote file \"%s\" on nodes %s with options %s"
                                     % (remote_file, list(self.nodes.keys()), kwargs))
            # This is a W/A to avoid https://github.com/paramiko/paramiko/issues/1133
            # if text contains non-ASCII characters.
            # Use the same encoding as paramiko uses, see paramiko/file.py/BufferedFile.write()
            bytes_stream = io.BytesIO(local_file.getvalue().encode('utf-8'))
            self._put(bytes_stream, remote_file, **kwargs)
            return

        self.cluster.log.verbose("Local file \"%s\" is being transferred to remote file \"%s\" on nodes %s with options %s"
                                 % (local_file, remote_file, list(self.nodes.keys()), kwargs))

        self.cluster.log.verbose('File size: %s' % os.path.getsize(local_file))
        local_file_hash = self.get_local_file_sha1(local_file)
        self.cluster.log.verbose('Local file hash: %s' % local_file_hash)
        remote_file_hashes = self.get_remote_file_sha1(remote_file)
        self.cluster.log.verbose('Remote file hashes: %s' % remote_file_hashes)

        hosts_to_upload = []
        for remote_ip, remote_file_hash in remote_file_hashes.items():
            if remote_file_hash != local_file_hash:
                self.cluster.log.verbose('Local and remote hashes does not match on node \'%s\' %s %s'
                                         % (remote_ip, local_file_hash, remote_file_hash))
                hosts_to_upload.append(remote_ip)
        if not hosts_to_upload:
            self.cluster.log.verbose('Local and remote hashes are equal on all nodes, no transmission required')
            return

        group_to_upload = self.cluster.make_group(hosts_to_upload)

        with open(local_file, "rb") as local_stream:
            group_to_upload._put(local_stream, remote_file, **kwargs)

    def _put(self, local_stream: IO, remote_file: str, **kwargs):
        hide = kwargs.pop("hide", True) is True
        sudo = kwargs.pop("sudo", False) is True
        backup = kwargs.pop("backup", False) is True
        mkdir = kwargs.pop("mkdir", False) is True
        immutable = kwargs.pop("immutable", False) is True

        # for unknown reason fabric v2 can't put async
        # Let's remember passed value, which by default is True, and make it False forcibly.
        is_async = kwargs.pop("is_async", True) is not False
        kwargs["is_async"] = False

        if sudo:
            self.cluster.log.verbose('A sudoer upload required')

        if backup:
            self.cluster.log.verbose('File \"%s\" backup required' % remote_file)

        if mkdir:
            self.cluster.log.verbose('A parent directory will be created')

        if immutable:
            self.cluster.log.verbose('File \"%s\" immutable set required' % remote_file)

        if not sudo and not backup and not immutable:
            # no additional commands execution is required - directly upload file
            self.do("put", local_stream, remote_file, **kwargs)
            return

        # for unknown reason fabric v2 can't put as sudo, and we should use WA via mv
        # also, if we need to backup the file first, then we also have to upload file to tmp first

        temp_filepath = "/tmp/%s" % uuid.uuid4().hex
        self.cluster.log.verbose("Uploading to temporary file '%s'..." % temp_filepath)
        self.do("put", local_stream, temp_filepath, **kwargs)

        self.cluster.log.verbose("Moving temporary file '%s' to '%s'..." % (temp_filepath, remote_file))

        if sudo:
            mv_command = "sudo chown root:root %s && sudo mv -f %s %s" % (temp_filepath, temp_filepath, remote_file)
        else:
            mv_command = "mv -f %s %s" % (temp_filepath, remote_file)

        if backup:
            if sudo:
                mv_command = "sudo cp -f %s %s.bak$(sudo ls %s* | sudo wc -l); %s" \
                             % (remote_file, remote_file, remote_file, mv_command)
            else:
                mv_command = "cp -f %s %s.bak$(ls %s* | wc -l); %s" \
                             % (remote_file, remote_file, remote_file, mv_command)

        if mkdir:
            file_directory = "/".join(remote_file.split('/')[:-1])
            if sudo:
                mv_command = f"sudo mkdir -p {file_directory}; {mv_command}"
            else:
                mv_command = f"mkdir -p {file_directory}; {mv_command}"

        mv_command = "cmp --silent %s %s || (%s)" % (remote_file, temp_filepath, mv_command)

        if immutable:
            if sudo:
                mv_command = "sudo chattr -i %s; %s; sudo chattr +i %s" % (remote_file, mv_command, remote_file)
            else:
                mv_command = "chattr -i %s; %s; chattr +i %s" % (remote_file, mv_command, remote_file)

        kwargs["hide"] = hide
        kwargs["is_async"] = is_async
        self.sudo(mv_command, **kwargs)

    def get(self, *args, **kwargs):
        return self.do("get", *args, **kwargs)

    def do(self, do_type, *args, **kwargs) -> Union[NodeGroupResult, int]:
        raw_results = self._do_with_wa(do_type, *args, **kwargs)
        if isinstance(raw_results, int):
            return raw_results
        group_results = self._make_result_or_fail(raw_results, lambda host, result: isinstance(result, Exception))

        if not kwargs.get('hide', True):
            self.cluster.log.debug(group_results, extra={'ignore_stdout': True})

        return group_results

    def _do_with_wa(self, do_type, *args, **kwargs) -> Union[_HostToResult, int]:
        # by default all code is async, but can be set False forcibly
        is_async = kwargs.pop("is_async", True) is not False

        left_nodes = self.nodes
        retry = 0
        results: _HostToResult = {}
        while True:
            retry += 1

            result = self._do(do_type, left_nodes, is_async, *args, **kwargs)
            if isinstance(result, int):
                return result

            results.update(result)
            left_nodes = {host: left_nodes[host] for host, result in results.items() if isinstance(result, Exception)}

            if not left_nodes or retry >= self.cluster.globals['workaround']['retries'] \
                    or not self._try_workaround(results, left_nodes):
                break

            self.cluster.log.verbose('Retrying #%s...' % retry)
            time.sleep(self.cluster.globals['workaround']['delay_period'])

        return results

    def _try_workaround(self, results: _HostToResult, failed_nodes: Connections) -> bool:
        not_booted = []

        for host in failed_nodes.keys():
            exception = results[host]
            if isinstance(exception, UnexpectedExit):
                exception_message = str(exception.result)
            else:
                exception_message = str(exception)

            if self.is_allowed_etcd_exception(exception_message):
                self.cluster.log.verbose("Detected ETCD problem at %s, need retry: %s" % (host, exception_message))
            elif self.is_allowed_kubernetes_exception(exception_message):
                self.cluster.log.verbose("Detected kubernetes problem at %s, need retry: %s" % (host, exception_message))
            elif self.is_allowed_connection_exception(exception_message):
                self.cluster.log.verbose("Detected connection exception at %s, will try to reconnect to node. Exception: %s"
                                         % (host, exception_message))
                not_booted.append(host)
            else:
                self.cluster.log.verbose("Detected unavoidable exception at %s, trying to solve automatically: %s"
                                         % (host, exception_message))
                return False

        # if there are not booted nodes, but we succeeded to wait for at least one is booted, we can continue execution
        if not_booted and self.cluster.make_group(not_booted).wait_and_get_active_nodes().is_empty():
            return False

        return True

    def _do(self, do_type: str, nodes: Connections, is_async, *args, **kwargs) -> Union[_HostToResult, int]:

        if do_type in ["run", "sudo"]:
            # by default fabric will print all output from nodes
            # let's disable this feature if it was not forcibly defined
            if kwargs.get("hide") is None:
                kwargs['hide'] = True

            if kwargs.get("timeout", None) is None:
                kwargs["timeout"] = self.cluster.globals['nodes']['command_execution']['timeout']

        execution_timeout = kwargs.get("timeout", None)

        results = {}

        if not nodes:
            self.cluster.log.verbose('No nodes to perform %s %s with options: %s' % (do_type, args, kwargs))
            return results

        self.cluster.log.verbose('Performing %s %s on nodes %s with options: %s' % (do_type, args, list(nodes.keys()), kwargs))

        executor = RemoteExecutor(self.cluster, lazy=False, parallel=is_async, timeout=execution_timeout)
        results = executor.queue(nodes, (do_type, args, kwargs))

        if not isinstance(results, int):
            simplified_results = {}
            for cnx, conn_results in results.items():
                raw_results = list(conn_results.values())
                if len(raw_results) > 1:
                    raise Exception('Unexpected condition: not supported multiple results with non-lazy GRE')
                simplified_results[cnx.host] = raw_results[0]
            return simplified_results

        return results

    def call(self, action, **kwargs):
        return self.call_batch([action], **{"%s.%s" % (action.__module__, action.__name__): kwargs})

    def call_batch(self, actions, **kwargs):
        results = {}

        for action in actions:

            callable_path = "%s.%s" % (action.__module__, action.__name__)
            self.cluster.log.debug("Running %s: " % callable_path)

            action_kwargs = {}
            if kwargs.get(callable_path) is not None:
                action_kwargs = kwargs[callable_path]

            results[action] = action(self, **action_kwargs)
            if results[action] is not None:
                self.cluster.log.debug(results[action])

        return results

    def get_online_nodes(self, online: bool) -> 'NodeGroup':
        online = [host for host, node_context in self.cluster.context['nodes'].items()
                  if node_context['access']['online'] == online]
        return self.cluster.make_group(online).intersection_group(self)

    def get_accessible_nodes(self) -> 'NodeGroup':
        accessible = [host for host, node_context in self.cluster.context['nodes'].items()
                      if node_context['access']['accessible']]
        return self.cluster.make_group(accessible).intersection_group(self)

    def get_sudo_nodes(self) -> 'NodeGroup':
        sudo = [host for host, node_context in self.cluster.context['nodes'].items()
                if node_context['access']['sudo'] != "No"]
        return self.cluster.make_group(sudo).intersection_group(self)

    def wait_for_reboot(self, initial_boot_history: NodeGroupResult, timeout=None) -> NodeGroupResult:
        results = self._await_rebooted_nodes(timeout, initial_boot_history=initial_boot_history)
        return self._make_result_or_fail(
            results,
            lambda host, result: isinstance(result, Exception) or result == initial_boot_history.get(self.nodes[host])
        )

    def wait_and_get_boot_history(self, timeout=None) -> NodeGroupResult:
        results = self._await_rebooted_nodes(timeout)
        return self._make_result_or_fail(results, lambda _, r: isinstance(r, Exception))

    def wait_and_get_active_nodes(self, timeout=None) -> 'NodeGroup':
        results = self._await_rebooted_nodes(timeout)
        not_booted = [host for host, result in results.items() if isinstance(result, Exception)]
        return self.exclude_group(self.cluster.make_group(not_booted))

    def _await_rebooted_nodes(self, timeout=None, initial_boot_history: NodeGroupResult = None) -> _HostToResult:

        if timeout is None:
            timeout = int(self.cluster.inventory['globals']['nodes']['boot']['timeout'])

        delay_period = self.cluster.globals['nodes']['boot']['defaults']['delay_period']

        if initial_boot_history:
            self.cluster.log.verbose("Initial boot history:\n%s" % initial_boot_history)
        else:
            initial_boot_history = NodeGroupResult(self.cluster)

        left_nodes = self.nodes
        results: _HostToResult = {}
        time_start = datetime.now()

        self.cluster.log.verbose("Trying to connect to nodes, timeout is %s seconds..." % timeout)

        # each connection has timeout, so the only we need is to repeat connecting attempts
        # during specified number of seconds
        while True:
            attempt_time_start = datetime.now()
            self.disconnect(list(left_nodes.keys()))

            self.cluster.log.verbose("Attempting to connect to nodes...")
            # this should be invoked without explicit timeout, and relied on fabric Connection timeout instead.
            results.update(self._do_nopasswd(left_nodes, "last reboot"))
            left_nodes = {host: left_nodes[host] for host, result in results.items()
                          if (isinstance(result, Exception)
                              # Something is wrong with sudo access. Node is active.
                              and not NodeGroup.is_require_nopasswd_exception(result))
                          or (not isinstance(result, Exception)
                              and result == initial_boot_history.get(self.nodes[host]))}

            waited = (datetime.now() - time_start).total_seconds()

            if not left_nodes or waited >= timeout:
                break

            for host, exc in results.items():
                if isinstance(exc, Exception) and not self.is_allowed_connection_exception(str(exc)):
                    self.cluster.log.verbose("Unexpected exception at %s, node is considered as not booted: %s"
                                             % (host, str(exc)))

            self.cluster.log.verbose("Nodes %s are not ready yet, remaining time to wait %i"
                                     % (list(left_nodes.keys()), timeout - waited))

            attempt_time = (datetime.now() - attempt_time_start).total_seconds()
            if attempt_time < delay_period:
                time.sleep(delay_period - attempt_time)

        if left_nodes:
            self.cluster.log.verbose("Failed to wait for boot of nodes %s" % list(left_nodes.keys()))
        else:
            self.cluster.log.verbose("All nodes are online now")

        return results

    def _do_nopasswd(self, left_nodes: Connections, command: str):
        prompt = '[sudo] password: '

        class NoPasswdResponder(invoke.Responder):
            def __init__(self):
                super().__init__(re.escape(prompt), None)

            def submit(self, stream):
                if self.pattern_matches(stream, self.pattern, "index"):
                    # If user appears to be not a NOPASSWD sudoer, "sudo" suggests to write password.
                    # This is a W/A to handle the situation in a docker container without pseudo-TTY (no -t option)
                    # As long as we require NOPASSWD, we can just fail immediately in such cases.
                    raise invoke.exceptions.ResponseNotAccepted("The user should be a NOPASSWD sudoer")

                # The only acceptable situation, responder does nothing.
                return []

        # Currently only NOPASSWD sudoers are supported.
        # Thus, running of connection.sudo("something") should be equal to connection.run("sudo something")
        return self._do("run", left_nodes, True, f"sudo -S -p '{prompt}' {command}",
                        watchers=[NoPasswdResponder()])

    @staticmethod
    def is_require_nopasswd_exception(exc: Exception):
        return isinstance(exc, invoke.exceptions.Failure) \
               and isinstance(exc.reason, invoke.exceptions.ResponseNotAccepted)

    def is_allowed_connection_exception(self, exception_message):
        exception_message = exception_message.partition('\n')[0]
        for known_exception_message in self.cluster.globals['connection']['bad_connection_exceptions']:
            if known_exception_message in exception_message:
                return True

        return False

    def is_allowed_etcd_exception(self, exception_message):
        for known_exception_message in self.cluster.globals['etcd']['temporary_exceptions']:
            if known_exception_message in exception_message:
                return True

        return False

    def is_allowed_kubernetes_exception(self, exception_message):
        for known_exception_message in self.cluster.globals['kubernetes']['temporary_exceptions']:
            if known_exception_message in exception_message:
                return True

        return False

    def get_local_file_sha1(self, filename):
        sha1 = hashlib.sha1()

        # Read local file by chunks of 2^16 bytes (65536) and calculate aggregated SHA1
        with open(filename, 'rb') as f:
            while True:
                data = f.read(2 ** 16)
                if not data:
                    break
                sha1.update(data)

        return sha1.hexdigest()

    def get_remote_file_sha1(self, filename):
        results = self._do_with_wa("sudo", "openssl sha1 %s" % filename, warn=True)
        self._make_result_or_fail(results, lambda h, r: isinstance(r, Exception))

        return {host: result.stdout.split("= ")[1].strip() if result.stdout else None
                for host, result in results.items()}

    def get_ordered_members_list(self, provide_node_configs=False, apply_filter=None) \
            -> List[Union[dict, 'NodeGroup']]:

        if apply_filter is None:
            apply_filter = {}

        result = []
        # we have to iterate strictly in order which was defined by user in config-file
        for node in self.cluster.inventory['nodes']:
            # is iterable node from inventory is part of current NodeGroup?
            if node['connect_to'] in self.nodes.keys():

                # apply filters
                suitable = True
                if apply_filter is not None:
                    if callable(apply_filter):
                        if not apply_filter(node):
                            suitable = False
                    else:
                        # here intentionally there is no way to filter by values in lists field,
                        # for this you need to use custom functions.
                        # Current solution implemented in this way because the filtering strategy is
                        # unclear - do I need to include when everything matches or is partial partial matching enough?
                        for key, value in apply_filter.items():
                            if node.get(key) is None:
                                suitable = False
                                break
                            if isinstance(value, list):
                                if node[key] not in value:
                                    suitable = False
                                    break
                            # elif should definitely be here, not if
                            elif node[key] != value:
                                suitable = False
                                break

                # if not filtered
                if suitable:
                    if provide_node_configs:
                        result.append(node)
                    else:
                        result.append(node['connection'])

        return result

    def get_member(self, number, provide_node_configs=False, apply_filter=None):
        results = self.get_ordered_members_list(provide_node_configs=provide_node_configs, apply_filter=apply_filter)

        if not results:
            return None

        return results[number]

    def get_first_member(self, provide_node_configs=False, apply_filter=None):
        return self.get_member(0, provide_node_configs=provide_node_configs, apply_filter=apply_filter)

    def get_last_member(self, provide_node_configs=False, apply_filter=None):
        return self.get_member(-1, provide_node_configs=provide_node_configs, apply_filter=apply_filter)

    def get_any_member(self, provide_node_configs=False, apply_filter=None):
        member = random.choice(self.get_ordered_members_list(provide_node_configs=provide_node_configs,
                                                             apply_filter=apply_filter))
        if isinstance(member, NodeGroup):
            # to avoid "Selected node <kubemarine.core.group.NodeGroup object at 0x7f925625d070>" writing to log,
            # let's get node ip from selected member and pass to it to log
            member_str = str(list(member.nodes.keys())[0])
        else:
            member_str = str(member)
        self.cluster.log.verbose(f'Selected node {member_str}')
        return member

    def get_member_by_name(self, name, provide_node_configs=False):
        return self.get_first_member(provide_node_configs=provide_node_configs, apply_filter={"name": name})

    def new_group(self, apply_filter=None):
        return self.cluster.make_group(self.get_ordered_members_list(apply_filter=apply_filter))

    def include_group(self, group):
        if group is None:
            return self

        ips = list(self.nodes.keys()) + list(group.nodes.keys())
        return self.cluster.make_group(list(dict.fromkeys(ips)))

    def exclude_group(self, group):
        if group is None:
            return self

        ips = list(set(self.nodes.keys()) - set(group.nodes.keys()))
        return self.cluster.make_group(list(dict.fromkeys(ips)))

    def intersection_group(self, group):
        if group is None:
            return self.cluster.make_group([])

        ips = list(set(self.nodes.keys()).intersection(set(group.nodes.keys())))
        return self.cluster.make_group(list(dict.fromkeys(ips)))

    def disconnect(self, hosts: List[str] = None):
        for host, cxn in self.nodes.items():
            if host in (hosts or self.nodes.keys()):
                self.cluster.log.verbose('Disconnected session with %s' % host)
                cxn.close()
                cxn._sftp = None

    def get_nodes_names(self) -> List[str]:
        result = []
        members = self.get_ordered_members_list(provide_node_configs=True)
        for node in members:
            result.append(node['name'])
        return result

    def get_node_name(self) -> str:
        if len(self.nodes) != 1:
            raise Exception("Cannot get the only name from not a single node")

        return self.get_first_member(provide_node_configs=True)['name']

    def get_hosts(self) -> List[str]:
        members = self.get_ordered_members_list(provide_node_configs=True)
        return [node['connect_to'] for node in members]

    def get_host(self):
        if len(self.nodes) != 1:
            raise Exception("Cannot get the only host from not a single node")

        return list(self.nodes.keys())[0]

    def is_empty(self) -> bool:
        return not self.nodes

    def has_node(self, node_name):
        return self.get_first_member(apply_filter={"name": node_name}) is not None

    def get_new_nodes(self) -> NodeGroup:
        return self.intersection_group(self.cluster.nodes.get('add_node'))

    def get_new_nodes_or_self(self) -> NodeGroup:
        new_nodes = self.get_new_nodes()
        if not new_nodes.is_empty():
            return new_nodes
        return self

    def get_nodes_for_removal(self) -> NodeGroup:
        return self.intersection_group(self.cluster.nodes.get('remove_node'))

    def get_nodes_for_removal_or_self(self) -> NodeGroup:
        nodes_for_removal = self.get_nodes_for_removal()
        if not nodes_for_removal.is_empty():
            return nodes_for_removal
        return self

    def get_changed_nodes(self) -> NodeGroup:
        return self.get_new_nodes().include_group(self.get_nodes_for_removal())

    def get_unchanged_nodes(self) -> NodeGroup:
        return self.exclude_group(self.get_changed_nodes())

    def get_final_nodes(self) -> NodeGroup:
        return self.exclude_group(self.cluster.nodes.get('remove_node'))

    def get_initial_nodes(self) -> NodeGroup:
        return self.exclude_group(self.cluster.nodes.get('add_node'))

    def nodes_amount(self) -> int:
        """
        Returns the number of nodes within a group
        :return: Integer
        """
        return len(self.nodes.keys())

    def get_nodes_os(self) -> str:
        """
        Returns the detected operating system family for group.

        :return: Detected OS family, possible values: "debian", "rhel", "rhel8", "multiple", "unknown", "unsupported".
        """
        return self.cluster.get_os_family_for_nodes(self.nodes.keys())

    def is_multi_os(self) -> bool:
        """
        Returns true if same group contains nodes with multiple OS families
        :return: Boolean
        """
        return self.get_nodes_os() == 'multiple'

    def get_subgroup_with_os(self, os_family: str) -> NodeGroup:
        """
        Forms and returns a new group from the nodes of the original group that have a specific OS family
        :param os_family: The name of required OS family
        :return: NodeGroup
        """
        if os_family not in ['debian', 'rhel', 'rhel8']:
            raise Exception('Unsupported OS family provided')
        node_names = []
        for node in self.get_ordered_members_list(provide_node_configs=True):
            node_os_family = self.cluster.get_os_family_for_node(node['connect_to'])
            if node_os_family == os_family:
                node_names.append(node['name'])
        return self.cluster.make_group_from_nodes(node_names)



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from textwrap import dedent

from kubemarine.core.action import Action
from kubemarine.core.patch import Patch
from kubemarine.core.resources import DynamicResources
from kubemarine import plugins


class TheAction(Action):
    def __init__(self):
        super().__init__("Correct ingress-nginx on kubernetes v1.25.x")

    def run(self, res: DynamicResources):
        cluster = res.cluster()

        version = cluster.inventory['services']['kubeadm']['kubernetesVersion']
        if '.'.join(version.split('.')[:-1]) != 'v1.25':
            cluster.log.info(f"Patch is not relevant for kubernetes {version}")
            return

        nginx_ingress_plugin = cluster.inventory['plugins']['nginx-ingress-controller']
        if not nginx_ingress_plugin.get('install', False) \
                or nginx_ingress_plugin.get('installation', {}).get('procedures') is None:
            cluster.log.info("nginx-ingress-controller plugin is disabled or its procedures aren't defined")
            return

        cluster.log.info(f"The following plugins will be reinstalled: nginx-ingress-controller")
        plugins.install_plugin(cluster, 'nginx-ingress-controller', nginx_ingress_plugin['installation']['procedures'])



class FixNginxIngress(Patch):
    def __init__(self):
        super().__init__("correct_ingress_nginx")

    @property
    def action(self) -> Action:
        return TheAction()

    @property
    def description(self) -> str:
        return dedent(
            f"""\
            Reinstall nginx-ingress-controller plugin for kubernetes clusters on v1.25.X using corrected template.
            """.rstrip()
        )



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from textwrap import dedent

from kubemarine.core import flow
from kubemarine.core.action import Action
from kubemarine.core.patch import Patch
from kubemarine.core.resources import DynamicResources
from kubemarine.procedures import install


class TheAction(Action):
    def __init__(self):
        super().__init__("Correct haproxy user")

    def run(self, res: DynamicResources):
        cluster = res.cluster()
        if not cluster.nodes.get('balancer'):
            cluster.log.info("Skip no balancer nodes.")
            return

        cluster.log.info(f"The following loadbalancer will be configure: haproxy")
        flow.run_tasks(res, install.tasks, cumulative_points=install.cumulative_points,
                       tasks_filter=['deploy.loadbalancer.haproxy.configure'])



class FixHaproxy(Patch):
    def __init__(self):
        super().__init__("correct_haproxy_user")

    @property
    def action(self) -> Action:
        return TheAction()

    @property
    def description(self) -> str:
        return dedent(
            f"""\
            Reconfigure loadbalancer haproxy for kubernetes clusters using corrected template.
            """.rstrip()
        )


# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time


def override_priviledged_ports(cluster, service=None, namespace=None):
    cluster.log.debug('Unlocking privileged ports...')
    control_planes = cluster.nodes['control-plane']
    control_planes.sudo('sed \'/- kube-apiserver/a\    - --service-node-port-range=80-32000\' -i /etc/kubernetes/manifests/kube-apiserver.yaml', hide=False)
    control_planes.sudo('systemctl restart kubelet.service', hide=False)
    # TODO: Get rid of hardcoded timeout - Wait for service start on all nodes
    time.sleep(60)
    control_planes.get_first_member().sudo('kubectl patch svc %s -n %s -p \'[ { "op": "replace", "path": "/spec/ports/1/nodePort", "value": 443 }, { "op": "replace", "path": "/spec/ports/0/nodePort", "value": 80 } ]\' --type=\'json\'' % (service, namespace), hide=False)
    control_planes.sudo('sed \'/service-node-port-range=.*/d\' -i /etc/kubernetes/manifests/kube-apiserver.yaml', hide=False)
    control_planes.sudo('systemctl restart kubelet.service', hide=False)
    time.sleep(60)
